import logging
from d3m.metadata.problem import TaskKeyword
from alphad3m.pipeline_synthesis.d3mpipeline_builder import BaseBuilder
from alphad3m.data_ingestion.data_profiler import get_privileged_data
from alphad3m.utils import load_primitives_types

logger = logging.getLogger(__name__)

TEMPLATES = {
    'CLASSIFICATION':
        [
            'd3m.primitives.classification.random_forest.SKlearn',
            'd3m.primitives.classification.xgboost_gbtree.Common',
            'd3m.primitives.classification.extra_trees.SKlearn',
            'd3m.primitives.classification.gradient_boosting.SKlearn',
            'd3m.primitives.classification.ada_boost.SKlearn',
            'd3m.primitives.classification.linear_svc.SKlearn'
        ],
    'DEBUG_CLASSIFICATION':
        [
            'd3m.primitives.classification.random_forest.SKlearn',
            'd3m.primitives.classification.xgboost_gbtree.Common'
        ],
    'REGRESSION':
        [
            'd3m.primitives.regression.random_forest.SKlearn',
            'd3m.primitives.regression.xgboost_gbtree.Common',
            'd3m.primitives.regression.extra_trees.SKlearn',
            'd3m.primitives.regression.gradient_boosting.SKlearn',
            'd3m.primitives.regression.ada_boost.SKlearn',
            'd3m.primitives.regression.linear_svr.SKlearn'
        ],
    'DEBUG_REGRESSION':
        [
            'd3m.primitives.regression.random_forest.SKlearn',
            'd3m.primitives.regression.xgboost_gbtree.Common'
        ]
}


def generate_pipelines(task_keywords, dataset, problem, targets, features, hyperparameters, metadata, metrics, DBSession):
    # Primitives for LUPI problems are no longer available. So, just exclude privileged data
    privileged_data = get_privileged_data(problem, task_keywords)
    metadata['exclude_columns'] += privileged_data
    task_keywords_set = set(task_keywords)
    #  Verify if two sets have at-least one element common
    if task_keywords_set & {TaskKeyword.GRAPH_MATCHING, TaskKeyword.LINK_PREDICTION, TaskKeyword.VERTEX_NOMINATION,
                            TaskKeyword.VERTEX_CLASSIFICATION, TaskKeyword.CLUSTERING, TaskKeyword.OBJECT_DETECTION,
                            TaskKeyword.COMMUNITY_DETECTION, TaskKeyword.SEMISUPERVISED, TaskKeyword.LUPI}:
        template_name = 'DEBUG_CLASSIFICATION'
    elif task_keywords_set & {TaskKeyword.COLLABORATIVE_FILTERING, TaskKeyword.FORECASTING}:
        template_name = 'DEBUG_REGRESSION'
    elif TaskKeyword.REGRESSION in task_keywords_set:
        template_name = 'REGRESSION'
        if task_keywords_set & {TaskKeyword.IMAGE, TaskKeyword.TEXT, TaskKeyword.AUDIO, TaskKeyword.VIDEO}:
            template_name = 'DEBUG_REGRESSION'
    else:
        template_name = 'CLASSIFICATION'
        if task_keywords_set & {TaskKeyword.IMAGE, TaskKeyword.TEXT, TaskKeyword.AUDIO, TaskKeyword.VIDEO}:
            template_name = 'DEBUG_CLASSIFICATION'

    logger.info('Creating pipelines from template %s' % template_name)

    feature_types = metadata['only_attribute_types']
    preprocessing_primitives = ['d3m.primitives.data_cleaning.imputer.SKlearn']

    if 'http://schema.org/Text' in feature_types:
        preprocessing_primitives.append('d3m.primitives.feature_extraction.tfidf_vectorizer.SKlearn')
    if 'http://schema.org/DateTime' in feature_types:
        preprocessing_primitives.append('d3m.primitives.data_transformation.enrich_dates.DistilEnrichDates')
    if 'https://metadata.datadrivendiscovery.org/types/CategoricalData' in feature_types:
        preprocessing_primitives.append('d3m.primitives.data_transformation.encoder.DSBOX')
    if len(preprocessing_primitives) == 1:  # Encoders were not applied, so use to_numeric for all features
        preprocessing_primitives.append('d3m.primitives.data_transformation.to_numeric.DSBOX')

    estimator_primitives = TEMPLATES.get(template_name, [])
    include_primitives = hyperparameters.get('include_primitives', []) or []  # Use empty list when the value is None
    exclude_primitives = hyperparameters.get('exclude_primitives', []) or []  # Use empty list when the value is None
    preprocessing_primitives, estimator_primitives = modify_search_space(preprocessing_primitives, estimator_primitives,
                                                                         include_primitives, exclude_primitives)

    builder = BaseBuilder()
    pipeline_ids = []

    for estimator_primitive in estimator_primitives:
        template = preprocessing_primitives + [estimator_primitive]
        pipeline_id = builder.make_d3mpipeline(template, 'Template', dataset, None, targets, features, metadata,
                                               metrics, DBSession=DBSession)
        pipeline_ids.append(pipeline_id)

    return pipeline_ids


def modify_search_space(preprocessing_primitives, estimator_primitives, include_primitives, exclude_primitives):
    primitives_types = load_primitives_types()

    for exclude_primitive in exclude_primitives:
        primitive_type = primitives_types.get(exclude_primitive, None)
        if primitive_type in {'CLASSIFICATION', 'REGRESSION'}:
            estimator_primitives = [i for i in estimator_primitives if i != exclude_primitive]
        elif primitive_type is not None:
            preprocessing_primitives = [i for i in preprocessing_primitives if i != exclude_primitive]

    for include_primitive in include_primitives:
        primitive_type = primitives_types.get(include_primitive, None)
        if primitive_type in {'CLASSIFICATION', 'REGRESSION'}:
            if include_primitive not in estimator_primitives:
                estimator_primitives.append(include_primitive)
        elif primitive_type is not None:
            if include_primitive not in preprocessing_primitives:
                preprocessing_primitives.append(include_primitive)

    return preprocessing_primitives, estimator_primitives



import os
import logging
import itertools
from os.path import join, dirname
from nltk.grammar import Production, Nonterminal, CFG, is_terminal, is_nonterminal
from alphad3m.utils import load_primitives_types
from alphad3m.primitive_loader import load_primitives_hierarchy
from alphad3m.metalearning.grammar_builder import create_metalearningdb_grammar

logger = logging.getLogger(__name__)
BASE_GRAMMAR_PATH = join(dirname(__file__), 'resource/base_grammar.bnf')
COMPLETE_GRAMMAR_PATH = join(os.environ.get('D3MOUTPUTDIR'), 'temp', 'complete_grammar.bnf')
TASK_GRAMMAR_PATH = join(os.environ.get('D3MOUTPUTDIR'), 'temp', 'task_grammar.bnf')


def create_global_grammar(grammar_string, primitives):
    base_grammar = CFG.fromstring(grammar_string)
    new_productions = []

    for production in base_grammar.productions():
        primitive_type = production.lhs().symbol()
        if primitive_type in primitives:
            new_rhs_list = []
            for token in production.rhs():
                if isinstance(token, str) and token.startswith('primitive_'):
                    new_rhs_list.append(primitives[primitive_type])
                else:
                    new_rhs_list.append([token])
            for new_rhs in itertools.product(*new_rhs_list):
                new_productions.append(Production(production.lhs(), new_rhs))
        else:
            new_productions.append(production)

    complete_grammar = CFG(Nonterminal('S'), new_productions)

    with open(COMPLETE_GRAMMAR_PATH, 'w') as fout:
        fout.write('\n'.join([str(x) for x in complete_grammar.productions()]))

    return complete_grammar


def create_task_grammar(global_grammar, task):
    logger.info('Creating specific grammar for task %s' % task)
    productions = global_grammar.productions(Nonterminal(task))
    start_token = Nonterminal('S')
    new_productions = []

    for start_production in productions:
        first_token = start_production.rhs()[0]
        if is_nonterminal(first_token) and first_token.symbol().endswith('_TASK'):
            for new_start_production in global_grammar.productions(first_token):
                new_productions.append(Production(start_token, new_start_production.rhs()))
        else:
            new_productions.append(Production(start_token, start_production.rhs()))

    for production in global_grammar.productions():
        for new_production in new_productions:
            if production.lhs() in new_production.rhs() and production not in new_productions:
                new_productions.append(production)

    task_grammar = CFG(start_token, new_productions)

    with open(TASK_GRAMMAR_PATH, 'w') as fout:
        fout.write('\n'.join([str(x) for x in task_grammar.productions()]))

    return task_grammar


def create_game_grammar(grammar):
    # Convert a context-free grammar to the game format
    start_symbol = grammar.start().symbol()
    game_grammar = {'START': start_symbol, 'NON_TERMINALS': {}, 'TERMINALS': {}, 'RULES': {}, 'RULES_LOOKUP': {},
                    'RULES_PROBA': {'GLOBAL': {}, 'LOCAL': {}, 'TYPES': {}}}
    terminals = []

    logger.info('Creating game grammar')
    for production in grammar.productions():
        non_terminal = production.lhs().symbol()
        production_str = str(production).replace('\'', '')

        game_grammar['RULES'][production_str] = len(game_grammar['RULES']) + 1

        if non_terminal not in game_grammar['NON_TERMINALS']:
            game_grammar['NON_TERMINALS'][non_terminal] = len(game_grammar['NON_TERMINALS']) + 1

        if non_terminal not in game_grammar['RULES_LOOKUP']:
            game_grammar['RULES_LOOKUP'][non_terminal] = []
        game_grammar['RULES_LOOKUP'][non_terminal].append(production_str)

        for token in production.rhs():
            if is_terminal(token) and token != 'E' and token not in terminals:
                terminals.append(token)

    game_grammar['TERMINALS'] = {t: i+len(game_grammar['NON_TERMINALS']) for i, t in enumerate(terminals, 1)}
    game_grammar['TERMINALS']['E'] = 0  # Special case for the empty symbol

    return game_grammar


def add_probabilities(game_grammar, probabilities):
    for primitive_type, primitive_probabilities in probabilities['global'].items():
        for primitive, probability in primitive_probabilities.items():
            id_rule_str = '%s -> %s' % (primitive_type, primitive)
            if id_rule_str in game_grammar['RULES']:
                id_rule_int = game_grammar['RULES'][id_rule_str]
                game_grammar['RULES_PROBA']['GLOBAL'][id_rule_int] = (id_rule_str, probability)

    for pattern, pattern_probabilities in probabilities['local'].items():
        game_grammar['RULES_PROBA']['LOCAL'][pattern] = {}
        for primitive_type, primitive_probabilities in pattern_probabilities.items():
            for primitive, probability in primitive_probabilities.items():
                id_rule_str = '%s -> %s' % (primitive_type, primitive)
                if id_rule_str in game_grammar['RULES']:
                    id_rule_int = game_grammar['RULES'][id_rule_str]
                    game_grammar['RULES_PROBA']['LOCAL'][pattern][id_rule_int] = (id_rule_str, probability)
    game_grammar['RULES_PROBA']['TYPES'] = probabilities['types']

    return game_grammar


def modify_manual_grammar(encoders, use_imputer):
    new_grammar = ''

    with open(BASE_GRAMMAR_PATH) as fin:
        for production_rule in fin.readlines():
            if production_rule.startswith('ENCODERS -> '):
                if len(encoders) > 0:
                    production_rule = 'ENCODERS -> %s\n' % ' '.join(encoders)
                else:
                    production_rule = "ENCODERS -> 'E'\n"
            elif production_rule.startswith('IMPUTATION -> '):
                if not use_imputer:
                    production_rule = "IMPUTATION -> 'E'\n"
            new_grammar += production_rule

    return new_grammar


def modify_text_primitives(primitives, task_keywords):
    if 'text' not in task_keywords:
        # Ignore some text processing primitives for non-text tasks
        ignore_primitives = {'d3m.primitives.feature_extraction.count_vectorizer.SKlearn',
                             'd3m.primitives.feature_extraction.boc.UBC', 'd3m.primitives.feature_extraction.bow.UBC',
                             'd3m.primitives.feature_extraction.nk_sent2vec.Sent2Vec',
                             'd3m.primitives.feature_extraction.tfidf_vectorizer.BBNTfidfTransformer'}
        primitives['TEXT_FEATURIZER'] = [p for p in primitives['TEXT_FEATURIZER'] if p not in ignore_primitives]

    return primitives


def load_manual_grammar(task, task_keywords, encoders, use_imputer, include_primitives, exclude_primitives):
    primitives = load_primitives_hierarchy()
    primitives = modify_text_primitives(primitives, task_keywords)
    primitives = modify_search_space(primitives, include_primitives, exclude_primitives)
    grammar_string = modify_manual_grammar(encoders, use_imputer)
    global_grammar = create_global_grammar(grammar_string, primitives)
    task_grammar = create_task_grammar(global_grammar, task)
    game_grammar = create_game_grammar(task_grammar)

    return game_grammar


def load_automatic_grammar(task, dataset_path, target_column, task_keywords, include_primitives, exclude_primitives):
    grammar_string, primitives = create_metalearningdb_grammar(task, dataset_path, target_column, task_keywords)
    if grammar_string is None:
        return None

    primitives['hierarchy'] = modify_search_space(primitives['hierarchy'], include_primitives, exclude_primitives)
    global_grammar = create_global_grammar(grammar_string, primitives['hierarchy'])
    task_grammar = create_task_grammar(global_grammar, task)
    game_grammar = create_game_grammar(task_grammar)
    game_grammar = add_probabilities(game_grammar, primitives['probabilities'])

    return game_grammar


def modify_search_space(primitives, include_primitives, exclude_primitives):
    primitives_types = load_primitives_types()

    for exclude_primitive in exclude_primitives:
        primitive_type = primitives_types.get(exclude_primitive, None)
        if primitive_type in primitives:
            primitives[primitive_type] = [i for i in primitives[primitive_type] if i != exclude_primitive]

    for include_primitive in include_primitives:
        primitive_type = primitives_types.get(include_primitive, None)
        if primitive_type in primitives:
            if include_primitive not in primitives[primitive_type]:
                primitives[primitive_type].append(include_primitive)

    return primitives



import logging
import os
import json
import pkg_resources
import random
import d3m.metadata.base
import d3m.runtime
from sqlalchemy.orm import joinedload
from d3m.container import Dataset
from alphad3m.schema import database, convert
from alphad3m.utils import is_collection, get_dataset_sample
from d3m.metrics import class_map
from d3m.metadata.pipeline import Pipeline
from d3m.metadata.problem import PerformanceMetric, TaskKeyword

logger = logging.getLogger(__name__)


with pkg_resources.resource_stream('alphad3m', 'resource/pipelines/kfold_tabular_split.yaml') as fp:
    kfold_tabular_split = Pipeline.from_yaml(fp)

with pkg_resources.resource_stream('alphad3m', 'resource/pipelines/kfold_timeseries_split.yaml') as fp:
    kfold_timeseries_split = Pipeline.from_yaml(fp)

with pkg_resources.resource_stream('alphad3m', 'resource/pipelines/train-test-tabular-split.yaml') as fp:
    train_test_tabular_split = Pipeline.from_yaml(fp)

with pkg_resources.resource_stream('alphad3m', 'resource/pipelines/scoring.yaml') as fp:
    scoring_pipeline = Pipeline.from_yaml(fp)


def check_timeindicator(dataset_path):
    with open(dataset_path) as fin:
        dataset_doc = json.load(fin)

    columns = dataset_doc['dataResources'][0]['columns']
    timeindicator_index = None
    has_timeindicator = False
    for item in columns:
        if item['colType'] == 'dateTime':
            timeindicator_index = item['colIndex']
        if 'timeIndicator' in item['role']:
            has_timeindicator = True
            break

    if not has_timeindicator:
        dataset_doc['dataResources'][0]['columns'][timeindicator_index]['role'].append('timeIndicator')
        try:
            with open(dataset_path, 'w') as fout:
                json.dump(dataset_doc, fout, indent=4)
        except:
            logger.error('Saving timeIndicator on dataset')


@database.with_db
def score(pipeline_id, dataset_uri, sample_dataset_uri, metrics, problem, scoring_config, report_rank, msg_queue, db):
    # Get pipeline from database
    logger.info('About to score pipeline, id=%s, metrics=%s', pipeline_id, metrics)
    pipeline = (
        db.query(database.Pipeline)
            .filter(database.Pipeline.id == pipeline_id)
            .options(joinedload(database.Pipeline.modules),
                     joinedload(database.Pipeline.connections))
    ).one()

    scores = calculate_scores(pipeline, dataset_uri, sample_dataset_uri, metrics, problem, scoring_config)
    logger.info("Evaluation results:\n%s", scores)

    if scoring_config['method'] == 'RANKING':  # Only for TA2 evaluation
        report_rank = True

    save_scores(pipeline_id, scores, metrics, report_rank, db)


def calculate_scores(pipeline, dataset_uri, sample_dataset_uri, metrics, problem, scoring_config):
    dataset_uri_touse = dataset_uri

    if sample_dataset_uri:
        dataset_uri_touse = sample_dataset_uri
    if TaskKeyword.FORECASTING in problem['problem']['task_keywords']:
        check_timeindicator(dataset_uri_touse[7:])

    dataset = Dataset.load(dataset_uri_touse)
    scores = {}
    pipeline_split = None

    if TaskKeyword.FORECASTING in problem['problem']['task_keywords']:
        pipeline_split = kfold_timeseries_split

    elif scoring_config['method'] == 'K_FOLD':
        pipeline_split = kfold_tabular_split

    elif scoring_config['method'] == 'HOLDOUT':
        pipeline_split = train_test_tabular_split

    elif scoring_config['method'] == 'RANKING':
        pipeline_split = kfold_tabular_split
    else:
        logger.warning('Unknown evaluation method, using K_FOLD')
        pipeline_split = kfold_tabular_split

    # FIXME: Splitting pipeline fails when works with F1 and semisupervised task, so use F1_MACRO instead
    # See https://gitlab.com/datadrivendiscovery/common-primitives/-/issues/92#note_520784899
    if metrics[0]['metric'] == PerformanceMetric.F1 and TaskKeyword.SEMISUPERVISED in problem['problem']['task_keywords']:
        new_metrics = [{'metric': PerformanceMetric.F1_MACRO}]
        scores = evaluate(pipeline, pipeline_split, dataset, new_metrics, problem, scoring_config, dataset_uri)
        scores = change_name_metric(scores, new_metrics, new_metric=metrics[0]['metric'].name)
    else:
        scores = evaluate(pipeline, pipeline_split, dataset, metrics, problem, scoring_config, dataset_uri)

    return scores


def save_scores(pipeline_id, scores, metrics, report_rank, db):
    scores_db = []

    if len(scores) > 0:  # It's a valid pipeline
        scores_db = add_scores_db(scores, scores_db)
        if report_rank:  # For TA2 only evaluation
            scores = create_rank_metric(scores, metrics)
            scores_db = add_scores_db(scores, scores_db)
            logger.info("Evaluation results for RANK metric: \n%s", scores)

    record_db = database.Evaluation(pipeline_id=pipeline_id, scores=scores_db)  # Store scores
    db.add(record_db)
    db.commit()


def evaluate(pipeline, data_pipeline, dataset, metrics, problem, scoring_config, dataset_uri):
    if is_collection(dataset_uri[7:]) or TaskKeyword.GRAPH in problem['problem']['task_keywords']:
        # Sampling in memory
        dataset = get_dataset_sample(dataset, problem)

    json_pipeline = convert.to_d3m_json(pipeline)

    logger.info("Pipeline to be scored:\n\t%s", '\n\t'.join(['step_%02d: %s' % (i, x['primitive']['python_path'])
                                                             for i, x in enumerate(json_pipeline['steps'])]))

    d3m_pipeline = Pipeline.from_json_structure(json_pipeline, )

    run_scores, run_results = d3m.runtime.evaluate(
        pipeline=d3m_pipeline,
        data_pipeline=data_pipeline,
        scoring_pipeline=scoring_pipeline,
        problem_description=problem,
        inputs=[dataset],
        data_params=scoring_config,
        metrics=metrics,
        volumes_dir=os.environ.get('D3MSTATICDIR', None),
        context=d3m.metadata.base.Context.TESTING,
        random_seed=0,
    )

    for result in run_results:
        if result.has_error():
            if TaskKeyword.GRAPH in problem['problem']['task_keywords']:
                # FIXME: Splitting pipeline fails for some graph datasets (e.g. 49_facebook). So, score the same dataset
                logger.warning('Evaluation failed for graph dataset, trying to score over the whole input dataset')
                return fit_score(d3m_pipeline, problem, dataset, dataset, metrics, scoring_config)
            else:
                raise RuntimeError(result.pipeline_run.status['message'])

    #save_pipeline_runs(run_results.pipeline_runs)
    combined_folds = d3m.runtime.combine_folds([fold for fold in run_scores])
    scores = {}

    for _, row in combined_folds.iterrows():
        if row['fold'] not in scores:
            scores[row['fold']] = {}
        scores[row['fold']][row['metric']] = row['value']

    return scores


def create_rank_metric(scores, metrics):
    scores_tmp = {}

    for fold, fold_scores in scores.items():
        scores_tmp[fold] = {}
        for metric, current_score in fold_scores.items():
            if metric == metrics[0]['metric'].name:
                new_score = (1.0 - metrics[0]['metric'].normalize(current_score)) + random.random() * 1.e-12
                scores_tmp[fold]['RANK'] = new_score

    return scores_tmp


def change_name_metric(scores, metrics, new_metric):
    scores_tmp = {}

    for fold, fold_scores in scores.items():
        scores_tmp[fold] = {}
        for metric, current_score in fold_scores.items():
            if metric == metrics[0]['metric'].name:
                scores_tmp[fold][new_metric] = current_score

    return scores_tmp


def add_scores_db(scores_dict, scores_db):
    for fold, fold_scores in scores_dict.items():
        for metric, value in fold_scores.items():
            scores_db.append(database.EvaluationScore(fold=fold, metric=metric, value=value))

    return scores_db


def save_pipeline_runs(pipelines_runs):
    for pipeline_run in pipelines_runs:
        save_run_path = os.path.join(os.environ['D3MOUTPUTDIR'], 'pipeline_runs',
                                     pipeline_run.to_json_structure()['id'] + '.yml')

        with open(save_run_path, 'w') as fin:
            pipeline_run.to_yaml(fin, indent=2)


def fit_score(d3m_pipeline, problem, dataset_train, dataset_test, metrics, scoring_config):
    fitted_pipeline, predictions, result = d3m.runtime.fit(
        pipeline=d3m_pipeline,
        problem_description=problem,
        inputs=[dataset_train],
        data_params=scoring_config,
        volumes_dir=os.environ.get('D3MSTATICDIR', None),
        context=d3m.metadata.base.Context.TESTING,
        random_seed=0
    )

    if result.has_error():
        raise RuntimeError(result.pipeline_run.status['message'])

    run_scores, _ = d3m.runtime.score(
        predictions, [dataset_test],
        scoring_pipeline=scoring_pipeline,
        problem_description=problem,
        metrics=metrics,
        predictions_random_seed=fitted_pipeline.random_seed,
        volumes_dir=fitted_pipeline.volumes_dir,
        scratch_dir=fitted_pipeline.scratch_dir,
        context=d3m.metadata.base.Context.TESTING,
        random_seed=0,
    )
    score = run_scores['value'][0]
    metric = run_scores['metric'][0]
    scores = {i: {metric: score} for i in range(int(scoring_config['number_of_folds']))}

    return scores



"""The SQLAlchemy models we use to persist data.
"""

import enum
import functools
import logging
from sqlalchemy import Column, ForeignKey, create_engine, func, not_, select, \
    inspect
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import column_property, relationship, sessionmaker
from sqlalchemy.sql import functions
from sqlalchemy.types import Binary, Boolean, DateTime, Enum, Float, Integer, \
    String

from alphad3m.schema.sql_uuid import UUID, UuidMixin


logger = logging.getLogger(__name__)


Base = declarative_base()


class Pipeline(UuidMixin, Base):
    __tablename__ = 'pipelines'

    origin = Column(String, nullable=False)
    dataset = Column(String, nullable=False)
    created_date = Column(DateTime, nullable=False,
                          server_default=functions.now())
    task = Column(String, nullable=True)
    parameters = relationship('PipelineParameter', lazy='joined')
    modules = relationship('PipelineModule')
    connections = relationship('PipelineConnection')
    runs = relationship('Run')

    def __eq__(self, other):
        return type(other) is Pipeline and other.id == self.id

    def __repr__(self):
        return '<Pipeline %r>' % self.id

    __str__ = __repr__


class PipelineModule(UuidMixin, Base):
    __tablename__ = 'pipeline_modules'

    pipeline_id = Column(UUID, ForeignKey('pipelines.id'), nullable=False)
    pipeline = relationship(Pipeline)
    package = Column(String, nullable=False)
    version = Column(String, nullable=False)
    name = Column(String, nullable=False)
    # connections_from = relationship('PipelineConnection')
    # connections_to = relationship('PipelineConnection')


class PipelineConnection(Base):
    __tablename__ = 'pipeline_connections'

    pipeline_id = Column(UUID, ForeignKey('pipelines.id'), nullable=False)
    pipeline = relationship(Pipeline)

    from_module_id = Column(UUID, ForeignKey('pipeline_modules.id'),
                            primary_key=True)
    from_module = relationship('PipelineModule',
                               foreign_keys=[from_module_id],
                               backref='connections_from')
    from_output_name = Column(String, primary_key=True)

    to_module_id = Column(UUID, ForeignKey('pipeline_modules.id'),
                          primary_key=True)
    to_module = relationship('PipelineModule',
                             foreign_keys=[to_module_id],
                             backref='connections_to')
    to_input_name = Column(String, primary_key=True)


class PipelineParameter(Base):
    __tablename__ = 'pipeline_parameters'

    pipeline_id = Column(UUID, ForeignKey('pipelines.id'), nullable=False)
    pipeline = relationship(Pipeline)

    module_id = Column(UUID, ForeignKey('pipeline_modules.id'),
                       primary_key=True)
    module = relationship('PipelineModule')
    name = Column(String, primary_key=True)
    value = Column(String, nullable=True)


class Evaluation(UuidMixin, Base):
    __tablename__ = 'evaluations'

    pipeline_id = Column(UUID, ForeignKey('pipelines.id'), nullable=False)
    pipeline = relationship('Pipeline')
    date = Column(DateTime, nullable=False,
                  server_default=functions.now())
    scores = relationship('EvaluationScore', lazy='joined')


class EvaluationScore(Base):
    __tablename__ = 'evaluation_scores'

    evaluation_id = Column(UUID, ForeignKey('evaluations.id'),
                                 primary_key=True)
    evaluation = relationship('Evaluation')
    fold = Column(Integer, primary_key=True, nullable=True)
    metric = Column(String, primary_key=True)
    value = Column(Float, nullable=False)


class RunType(enum.Enum):
    TRAIN = 1
    TEST = 2


class Run(UuidMixin, Base):
    __tablename__ = 'runs'

    pipeline_id = Column(UUID, ForeignKey('pipelines.id'), nullable=False)
    pipeline = relationship('Pipeline')
    date = Column(DateTime, nullable=False,
                  server_default=functions.now())
    reason = Column(String, nullable=False)
    type = Column(Enum(RunType), nullable=False)
    special = Column(Boolean, nullable=False, default=False)
    inputs = relationship('Input')
    outputs = relationship('Output')


class Input(Base):
    __tablename__ = 'inputs'

    run_id = Column(UUID, ForeignKey('runs.id'), primary_key=True)
    run = relationship('Run')
    module_id = Column(UUID, ForeignKey('pipeline_modules.id'),
                       primary_key=True)
    module = relationship('PipelineModule')
    input_name = Column(String, primary_key=True, nullable=True)
    value = Column(Binary, nullable=False)


class Output(Base):
    __tablename__ = 'outputs'

    run_id = Column(UUID, ForeignKey('runs.id'), primary_key=True)
    run = relationship('Run')
    module_id = Column(UUID, ForeignKey('pipeline_modules.id'),
                       primary_key=True)
    module = relationship('PipelineModule')
    output_name = Column(String, primary_key=True, nullable=True)
    value = Column(Binary, nullable=False)


# Trained true iff there's a Run with special=False and type=TRAIN
Pipeline.trained = column_property(
    select(
        [func.count(Run.id)]
    ).where((Run.pipeline_id == Pipeline.id) &
            (not_(Run.special)) &
            (Run.type == RunType.TRAIN))
    .as_scalar() != 0
)


def connect(filename):
    """Connect to the database using an environment variable.
    """
    logger.info("Connecting to SQL database")
    url = 'sqlite:///{0}'.format(filename)
    engine = create_engine(url, echo=False)

    if not engine.dialect.has_table(engine.connect(), 'pipelines'):
        logger.warning("The tables don't seem to exist; creating")
        Base.metadata.create_all(bind=engine)

    return engine, sessionmaker(bind=engine,
                                autocommit=False,
                                autoflush=False)


def with_db(wrapped):
    @functools.wraps(wrapped)
    def wrapper(*args, db_filename=None, **kwargs):
        engine, DBSession = connect(db_filename)
        db = DBSession()
        try:
            return wrapped(*args, **kwargs, db=db)
        finally:
            db.close()
    return wrapper


def with_sessionmaker(wrapped):
    @functools.wraps(wrapped)
    def wrapper(*args, db_filename=None, **kwargs):
        engine, DBSession = connect(db_filename)
        return wrapped(*args, **kwargs, DBSession=DBSession)
    return wrapper


def _duplicate(db, obj, replace={}):
    new_obj = type(obj)()
    mapper = inspect(type(obj))
    for k, col in mapper.columns.items():
        if k in replace:
            v = replace[k]
            if isinstance(v, dict):
                old_v = getattr(obj, k)
                v = v[old_v]
        elif k != 'id':
            v = getattr(obj, k)
        else:
            continue
        setattr(new_obj, k, v)
    db.add(new_obj)
    return new_obj


def _duplicate_list(db, lst, replace={}):
    return [_duplicate(db, obj, replace) for obj in lst]


def duplicate_pipeline(db, pipeline, origin):
    """Duplicates a Pipeline.

    This creates a new Pipeline with new modules/parameters/connections. They
    all get new IDs.
    """
    new_pipeline = Pipeline(
        origin=origin,
        task=pipeline.task,
        dataset=pipeline.dataset,
    )
    db.add(new_pipeline)

    modules = {}
    for module in pipeline.modules:
        new_module = _duplicate(db, module,
                                dict(pipeline_id=new_pipeline.id))
        modules[module.id] = new_module.id

    _duplicate_list(db, pipeline.parameters,
                    dict(pipeline_id=new_pipeline.id,
                         module_id=modules))
    _duplicate_list(db, pipeline.connections,
                    dict(pipeline_id=new_pipeline.id,
                         from_module_id=modules,
                         to_module_id=modules))

    db.flush()
    return new_pipeline



'''Entrypoints for AlphaD3M.

This contains the multiple entrypoints for AlphaD3M, that get called from
different commands. They spin up a AutoML object and use it.
'''

import logging
import os
import sys
from alphad3m.automl import AutoML


logger = logging.getLogger(__name__)


def setup_logging():
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s:%(levelname)s:AlphaD3M:%(name)s:%(message)s')


def main_serve():
    setup_logging()

    port = int(os.environ['D3MPORT'])
    output_folder = os.environ['D3MOUTPUTDIR']

    automl = AutoML(output_folder)
    automl.run_server(port)


def main_search():
    setup_logging()
    timeout = None
    output_folder = None

    if len(sys.argv) == 3:
        output_folder = sys.argv[1]
        timeout = int(sys.argv[2])

    dataset = '/input/TRAIN/dataset_TRAIN/datasetDoc.json'
    problem_path = '/input/TRAIN/problem_TRAIN/problemDoc.json'
    automl = AutoML(output_folder)
    automl.run_search(dataset, problem_path=problem_path, timeout=timeout)










import re
import os
import json
import logging
import datamart_profiler
import pandas as pd
from os.path import join, dirname
from d3m import index as d3m_index
from d3m.container import Dataset
from d3m.metadata.problem import TaskKeyword
from d3m.container.dataset import D3M_COLUMN_TYPE_CONSTANTS_TO_SEMANTIC_TYPES, D3M_ROLE_CONSTANTS_TO_SEMANTIC_TYPES
from alphad3m.utils import need_denormalize

logger = logging.getLogger(__name__)


def select_annotated_feature_types(dataset_doc_path):
    feature_types = {}

    with open(dataset_doc_path) as fin:
        dataset_doc = json.load(fin)

    try:
        for data_res in dataset_doc['dataResources']:
            if data_res['resID'] == 'learningData' and data_res['resType'] == 'table':
                for column in data_res['columns']:
                    if column['colType'] != 'unknown':
                        semantic_type = [D3M_COLUMN_TYPE_CONSTANTS_TO_SEMANTIC_TYPES[column['colType']]]
                        role = D3M_ROLE_CONSTANTS_TO_SEMANTIC_TYPES[column['role'][0]]
                        if 'refersTo' in column:  # It's a foreign key
                            role = 'https://metadata.datadrivendiscovery.org/types/PrimaryKey'
                        feature_types[column['colName']] = (role, semantic_type, column['colIndex'])
    except:
        logger.exception('Error reading the type of attributes')

    logger.info('Features with annotated types: [%s]', ', '.join(feature_types.keys()))

    return feature_types


def select_unkown_feature_types(csv_path, annotated_features):
    all_features = pd.read_csv(csv_path, index_col=0, nrows=0).columns
    unkown_feature_types = []

    for feature_name in all_features:
        match = re.match('(.*)\.\d$', feature_name)  # To verify if pandas renames features with same names
        if match:
            feature_name = match.group(1)
        if feature_name not in annotated_features:
            unkown_feature_types.append(feature_name)

    logger.info('Features with unknown types: [%s]', ', '.join(unkown_feature_types))

    return unkown_feature_types


def select_identified_feature_types(metadata, unkown_feature_types, target_names):
    identified_feature_types = {}

    for index, item in enumerate(metadata['columns']):
        feature_name = item['name']
        if feature_name in unkown_feature_types:
            semantic_types = item['semantic_types'] if len(item['semantic_types']) > 0 else [item['structural_type']]
            d3m_semantic_types = []
            for semantic_type in semantic_types:
                if semantic_type == 'http://schema.org/Enumeration':  # Changing to D3M format
                    semantic_type = 'https://metadata.datadrivendiscovery.org/types/CategoricalData'
                elif semantic_type == 'http://schema.org/identifier':  # Changing to D3M format
                    semantic_type = 'http://schema.org/Integer'
                elif semantic_type in {'http://schema.org/longitude', 'http://schema.org/latitude'}:
                    semantic_type = 'http://schema.org/Float'
                d3m_semantic_types.append(semantic_type)

            role = 'https://metadata.datadrivendiscovery.org/types/Attribute'
            if 'https://metadata.datadrivendiscovery.org/types/PrimaryKey' in d3m_semantic_types:
                role = 'https://metadata.datadrivendiscovery.org/types/PrimaryKey'
            elif feature_name in target_names:
                role = 'https://metadata.datadrivendiscovery.org/types/TrueTarget'
            identified_feature_types[feature_name] = (role, d3m_semantic_types, index)

    logger.info('Inferred feature types:\n%s',
                '\n'.join(['%s = [%s]' % (k, ', '.join([i for i in v[1]])) for k, v in identified_feature_types.items()]))

    return identified_feature_types


def get_extra_metadata(metadata, exclude_columns):
    extra_metadata = {'missing_values': False, 'large_rows': False, 'large_columns': False, 'exclude_columns': []}

    for index, item in enumerate(metadata['columns']):
        if 'missing_values_ratio' in item and item['name'] not in exclude_columns:
            extra_metadata['missing_values'] = True

        if 'structural_type' in item and 'https://metadata.datadrivendiscovery.org/types/MissingData' in item['structural_type']:
            extra_metadata['exclude_columns'].append(index)

    if 'nb_rows' in metadata and metadata['nb_rows'] > 1000000:
        extra_metadata['large_rows'] = True
        extra_metadata['sample_size'] = metadata['nb_profiled_rows']

    if 'nb_columns' in metadata and metadata['nb_columns'] > 1000:
        extra_metadata['large_columns'] = True

    return extra_metadata


def profile_data(dataset_uri, targets):
    dataset_doc_path = dataset_uri[7:]
    csv_path = join(dirname(dataset_doc_path), 'tables', 'learningData.csv')
    if need_denormalize(dataset_doc_path):
        csv_path = denormalize_dataset(dataset_uri)

    target_names = [x[1] for x in targets]
    annotated_feature_types = select_annotated_feature_types(dataset_doc_path)
    unkown_feature_types = select_unkown_feature_types(csv_path, annotated_feature_types.keys())
    metadata = datamart_profiler.process_dataset(csv_path, coverage=False)
    identified_feature_types = select_identified_feature_types(metadata, unkown_feature_types, target_names)
    only_attribute_types = set()
    semantictypes_by_index = {}

    for role, semantic_types, index in annotated_feature_types.values():
        if role == 'https://metadata.datadrivendiscovery.org/types/Attribute':
            only_attribute_types.update(semantic_types)

    for role, semantic_types, index in identified_feature_types.values():
        if role == 'https://metadata.datadrivendiscovery.org/types/Attribute':
            only_attribute_types.update(semantic_types)

        for semantic_type in set(semantic_types + [role]):
            if semantic_type not in semantictypes_by_index:
                semantictypes_by_index[semantic_type] = []
            semantictypes_by_index[semantic_type].append(index)

    features_metadata = {'semantictypes_indices': semantictypes_by_index, 'only_attribute_types': only_attribute_types}
    extra_metadata = get_extra_metadata(metadata, target_names)
    features_metadata.update(extra_metadata)

    return features_metadata


def denormalize_dataset(dataset_uri):
    logger.info('Denormalizing the dataset')
    csv_path = None
    dataset = Dataset.load(dataset_uri)

    try:
        primitive_class = d3m_index.get_primitive('d3m.primitives.data_transformation.denormalize.Common')
        primitive_hyperparams = primitive_class.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        primitive_denormalize = primitive_class(hyperparams=primitive_hyperparams.defaults())
        primitive_output = primitive_denormalize.produce(inputs=dataset).value

        primitive_class = d3m_index.get_primitive('d3m.primitives.data_transformation.dataset_to_dataframe.Common')
        primitive_hyperparams = primitive_class.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        primitive_dataframe = primitive_class(hyperparams=primitive_hyperparams.defaults())
        primitive_output = primitive_dataframe.produce(inputs=primitive_output).value

        csv_path = join(os.environ.get('D3MOUTPUTDIR'), 'temp', 'denormalized_dataset.csv')
        primitive_output.to_csv(csv_path)
    except:
        csv_path = join(dirname(dataset_uri[7:]),  'tables', 'learningData.csv')
        logger.exception('Error denormalizing dataset, using only learningData.csv file')

    return csv_path


def get_privileged_data(problem, task_keywords):
    privileged_data = []
    if TaskKeyword.LUPI in task_keywords and 'privileged_data' in problem['inputs'][0]:
        for column in problem['inputs'][0]['privileged_data']:
            privileged_data.append(column['column_index'])

    return privileged_data


def select_encoders(feature_types):
    encoders = []
    mapping_feature_types = {'https://metadata.datadrivendiscovery.org/types/CategoricalData': 'CATEGORICAL_ENCODER',
                             'http://schema.org/Text': 'TEXT_FEATURIZER', 'http://schema.org/DateTime': 'DATETIME_ENCODER'}

    for features_type in feature_types:
        if features_type in mapping_feature_types:
            encoders.append(mapping_feature_types[features_type])

    return sorted(encoders, reverse=True)



import json
import pickle
import logging
from os.path import join, dirname

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')
logger = logging.getLogger(__name__)

METALEARNINGDB_PATH = join(dirname(__file__), '../resource/metalearningdb.pkl')


def merge_pipeline_files(pipelines_file, pipeline_runs_file, problems_file, n=-1, verbose=False):
    logger.info('Adding pipelines to lookup table...')
    pipelines = {}
    with open(pipelines_file, 'r') as f:
        for line in f:
            pipeline = json.loads(line)
            pipelines[pipeline['digest']] = pipeline

    logger.info('Adding problems to lookup table...')
    problems = {}
    with open(problems_file, 'r') as f:
        for line in f:
            problem = json.loads(line)
            problems[problem['digest']] = problem['problem']
            problems[problem['digest']]['id'] = problem['id']

    logger.info('Merging pipeline information with pipeline_runs_file (this might take a while)...')
    merged = []
    with open(pipeline_runs_file, 'r') as f:
        for line in f:
            if len(merged) == n:
                break
            try:
                run = json.loads(line)
                if run['run']['phase'] != 'PRODUCE':
                    continue
                pipeline = pipelines[run['pipeline']['digest']]
                problem = problems[run['problem']['digest']]
                data = {
                    'pipeline_id': pipeline['id'],
                    'pipeline_digest': pipeline['digest'],
                    'pipeline_source': pipeline['source'],
                    'inputs': pipeline['inputs'],
                    'outputs': pipeline['outputs'],
                    'problem': problem,
                    'start': run['start'],
                    'end': run['end'],
                    'steps': pipeline['steps'],
                    'scores': run['run']['results']['scores']
                }
                merged.append(json.dumps(data))
            except Exception as e:
                if (verbose):
                    logger.error(problem['id'], repr(e))
    logger.info('Done.')

    with open(METALEARNINGDB_PATH, 'wb') as f:
        pickle.dump(merged, f)


def load_metalearningdb():
    all_pipelines = []
    logger.info('Loading pipelines from metalearning database...')
    with open(METALEARNINGDB_PATH, 'rb') as fin:
        # Use pickle instead of json because it was faster in our experiments
        all_pipelines = pickle.load(fin)

    logger.info('Found %d pipelines in metalearning database' % len(all_pipelines))

    return all_pipelines


if __name__ == '__main__':
    pipelines_file = '/Users/rlopez/Downloads/metalearningdb_dump_20200304/pipelines-1583354358.json'
    pipeline_runs_file = '/Users/rlopez/Downloads/metalearningdb_dump_20200304/pipeline_runs-1583354387.json'
    problems_file = '/Users/rlopez/Downloads/metalearningdb_dump_20200304/problems-1583354357.json'
    merge_pipeline_files(pipelines_file, pipeline_runs_file, problems_file)



from __future__ import print_function
import os
import pickle
import math
import logging
from copy import deepcopy

from alphad3m.pipeline_search.Game import Game
from alphad3m.pipeline_search.pipeline.PipelineLogic import Board
import numpy as np
import traceback
import time

logger = logging.getLogger(__name__)

class PipelineGame(Game):
    # FIXEME: Maybe the input parameters can be in json
    def __init__(self, input={}, eval_pipeline=None):
        self.steps = 0
        self.args = input['ARGS']
        self.evaluations = {}
        self.eval_times = {}

        self.grammar = input['GRAMMAR']
        self.pipeline_size = input['PIPELINE_SIZE']
        self.problem_types = input['PROBLEM_TYPES']
        self.data_types = input['DATA_TYPES']
        self.eval_pipeline = eval_pipeline

        self.problem = input['PROBLEM'].upper()
        self.data_type = input['DATA_TYPE'].upper()
        self.metric = input['METRIC']
        self.dataset = input['DATASET']

        self.dataset_metafeatures = input['DATASET_METAFEATURES']
        if self.dataset_metafeatures is None:
            metafeatures_path = args.get('metafeatures_path')
            if not metafeatures_path is None:
                metafeatures_file = os.path.join(metafeatures_path, args['dataset'] + '_metafeatures.pkl')
                if os.path.isfile(metafeatures_file):
                    m_f = open(metafeatures_file, 'rb')
                    self.dataset_metafeatures = pickle.load(m_f)[args['dataset']]

        if self.dataset_metafeatures is None:
            logger.warning('No Dataset Metafeatures specified - Initializing to empty')
            self.dataset_metafeatures = []
        else:
            self.dataset_metafeatures = list(np.nan_to_num(np.asarray(self.dataset_metafeatures)))
            
        self.m = len(self.dataset_metafeatures)+2
        self.p = input['PIPELINE_SIZE']
        self.action_size = 0

                
    def getInitBoard(self):
        # return initial board (numpy board)
        b = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        b.set_metafeatures(self.dataset_metafeatures+[self.data_types[self.data_type]]+[self.problem_types[self.problem]])
        self.action_size = len(b.valid_moves)
        return b.pieces_m + b.pieces_p

    def getBoardSize(self):
        # (a,b) tuple
        b = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        return b.get_board_size()

    def getActionSize(self):
        # return number of actions
        board = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        return len(board.valid_moves)

    def getNextState(self, board, player, action):
        # if player takes action on board, return next (board,player)
        # action must be a valid move
        b = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        b.set_metafeatures(board)
        b.set_pipeline(board)
        #logger.info('PREV STATE %s', b.pieces_p)
        b.execute_move(action, player)
        #logger.info('NEXT STATE %s', b.pieces_p)
        return (b.pieces_m+b.pieces_p, -player)

    def getValidMoves(self, board, player):
        # return a fixed size binary vector
        b = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        b.set_metafeatures(board)
        b.set_pipeline(board)
        #logger.info('CURR STATE %s', b.pieces_p)
        legalMoves =  b.get_legal_moves()
        #logger.info('VALID MOVES %s', [b.valid_moves[i] for i in range(0, len(legalMoves)) if legalMoves[i] == 1])
        return np.array(legalMoves)

    def getEvaluation(self, board):

        b = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        pipeline_enums = b.get_pipeline(board)
        if not any(pipeline_enums):
            return 0.0
        pipeline = b.get_pipeline_primitives(pipeline_enums)
        eval_val = self.evaluations.get(",".join(pipeline))

        if eval_val is None:
            self.steps = self.steps + 1
            try:
                eval_val = self.eval_pipeline(pipeline, 'AlphaAutoML')
            except:
                logger.warning('Error in Pipeline Execution %s', eval_val)
                traceback.print_exc()
            if eval_val is None:
                eval_val = float('inf')
            self.evaluations[",".join(pipeline)] = eval_val
            self.eval_times[",".join(pipeline)] = time.time()

        return eval_val
    
    def getGameEnded(self, board, player, eval_val=None):
        # return 0 if not ended, 1 if x won, -1 if x lost
        # player = 1

        b = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        b.set_metafeatures(board)
        b.set_pipeline(board)
        if not b.is_terminal_pipeline():
            return 0
        if len(self.evaluations) > 0:
            sorted_evals = sorted([eval for eval in list(self.evaluations.values()) if eval != float('inf')])
            if len(sorted_evals) > 0:
                if 'error' in self.metric.lower():
                   win_threshold = sorted_evals[0]
                else:
                   win_threshold = sorted_evals[-1]
                b.win_threshold = win_threshold

        eval_val = self.getEvaluation(board)

        if b.findWin(player, eval_val):
            logger.info('findwin %s', player)
            return 1
        if b.findWin(-player, eval_val):
            logger.info('findwin %', -player)
            return -1
        if b.has_legal_moves():
            return 0

        return 2

    def getCanonicalForm(self, board, player):
        # return state if player==1, else return -state if player==-1
        return deepcopy(board)

    def stringRepresentation(self, board):
        # 3x3 numpy array (canonical board)
        return np.asarray(board).tostring()

    def getTrainBoard(self, board):
        b = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        b.set_metafeatures(board)
        b.set_pipeline(board)
        return b.get_train_board()

    def getTrainExamples(self, board, pi):
        assert(len(pi) == self.getActionSize())  # 1 for pass
        b = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        b.set_metafeatures(board)
        b.set_pipeline(board)
        if not b.is_terminal_pipeline():
            eval_val = float('inf')
        else:
            eval_val = self.getEvaluation(board)
        train_board = b.get_train_board()
        if 'error' in self.metric.lower():
            return (train_board, pi, eval_val if eval_val!= float('inf') and eval_val <= math.pow(10,15) else math.pow(10,15))
        else:
            return (train_board, pi, eval_val if eval_val != float('inf') else 0)

    def get_pipeline_primitives(self, board):
        b = Board(self.m, self.grammar, self.pipeline_size, self.metric)
        return b.get_pipeline_primitives(b.get_pipeline(board))
        
    def display(self, b):
        board = self.get_pipeline_primitives(b)
        logger.info("PIPELINE: %s", '|'.join(e for e in board))




import os
import logging
import pickle
import frozendict
from alphad3m.schema import database
from d3m.container import Dataset
from d3m.metadata import base as metadata_base
from alphad3m.pipeline_operations.pipeline_execute import execute
logger = logging.getLogger(__name__)

METAFEATURES_DEFAULT = [
                        'dimensionality',
                        'number_distinct_values_of_categorical_attributes.max',
                        'number_distinct_values_of_numeric_attributes.max',
                        'entropy_of_categorical_attributes.max',
                        'kurtosis_of_attributes.max',
                        'mean_of_attributes.max',
                        'entropy_of_numeric_attributes.max',
                        'skew_of_attributes.max',
                        'standard_deviation_of_attributes.max',
                        'number_distinct_values_of_categorical_attributes.mean',
                        'number_distinct_values_of_numeric_attributes.mean',
                        'entropy_of_categorical_attributes.mean',
                        'kurtosis_of_attributes.mean',
                        'mean_of_attributes.mean',
                        'entropy_of_numeric_attributes.mean',
                        'skew_of_attributes.mean',
                        'standard_deviation_of_attributes.mean',
                        'number_distinct_values_of_categorical_attributes.min',
                        'number_distinct_values_of_numeric_attributes.min',
                        'entropy_of_categorical_attributes.min',
                        'kurtosis_of_attributes.min',
                        'mean_of_attributes.min',
                        'entropy_of_numeric_attributes.min',
                        'skew_of_attributes.min',
                        'standard_deviation_of_attributes.min',
                        'number_of_categorical_attributes',
                        'number_of_attributes',
                        'number_of_features_with_missing_values',
                        'number_of_instances',
                        'number_of_instances_with_missing_values',
                        'number_of_missing_values',
                        'number_of_numeric_attributes',
                        'pca.eigenvalue_component_1',
                        'pca.eigenvalue_component_2',
                        'pca.eigenvalue_component_3',
                        'pca.explained_variance_ratio_component_1',
                        'pca.explained_variance_ratio_component_2',
                        'pca.explained_variance_ratio_component_3',
                        'entropy_of_categorical_attributes.quartile_1',
                        'kurtosis_of_attributes.quartile_1',
                        'mean_of_attributes.quartile_1',
                        'entropy_of_numeric_attributes.quartile_1',
                        'skew_of_attributes.quartile_1',
                        'standard_deviation_of_attributes.quartile_1',
                        'entropy_of_categorical_attributes.median',
                        'kurtosis_of_attributes.median',
                        'mean_of_attributes.median',
                        'entropy_of_numeric_attributes.median',
                        'skew_of_attributes.median',
                        'standard_deviation_of_attributes.median',
                        'entropy_of_categorical_attributes.quartile_3',
                        'kurtosis_of_attributes.quartile_3',
                        'mean_of_attributes.quartile_3',
                        'entropy_of_numeric_attributes.quartile_3',
                        'skew_of_attributes.quartile_3',
                        'standard_deviation_of_attributes.quartile_3',
                        'ratio_of_categorical_attributes',
                        'ratio_of_features_with_missing_values',
                        'ratio_of_instances_with_missing_values',
                        'ratio_of_missing_values',
                        'ratio_of_numeric_attributes',
                        'number_distinct_values_of_categorical_attributes.std',
                        'number_distinct_values_of_numeric_attributes.std',
                        'kurtosis_of_attributes.std',
                        'mean_of_attributes.std',
                        'skew_of_attributes.std',
                        'standard_deviation_of_attributes.std'
                        ]


class ComputeMetafeatures():

    def __init__(self, dataset, targets=None, features=None, DBSession=None):
        self.dataset = Dataset.load(dataset)
        self.dataset_uri = dataset
        self.DBSession = DBSession
        self.targets = targets
        self.features = features

    def _add_target_columns_metadata(self):
        for target in self.targets:
            resource_id = target[0]
            target_name = target[1]
            for column_index in range(self.dataset.metadata.query((resource_id, metadata_base.ALL_ELEMENTS))['dimension']['length']):
                if self.dataset.metadata.query((resource_id, metadata_base.ALL_ELEMENTS, column_index)).get('name',
                                                                                                None) == target_name:
                    semantic_types = list(self.dataset.metadata.query(
                        (resource_id, metadata_base.ALL_ELEMENTS, column_index)).get('semantic_types', []))

                    if 'https://metadata.datadrivendiscovery.org/types/Target' not in semantic_types:
                        semantic_types.append('https://metadata.datadrivendiscovery.org/types/Target')
                        self.dataset.metadata = self.dataset.metadata.update(
                            (resource_id, metadata_base.ALL_ELEMENTS, column_index),
                            {'semantic_types': semantic_types})

                    if 'https://metadata.datadrivendiscovery.org/types/TrueTarget' not in semantic_types:
                        semantic_types.append('https://metadata.datadrivendiscovery.org/types/TrueTarget')
                        self.dataset.metadata = self.dataset.metadata.update(
                            (resource_id, metadata_base.ALL_ELEMENTS, column_index),
                            {'semantic_types': semantic_types})

    def _create_metafeatures_pipeline(self, db, origin):

        pipeline = database.Pipeline(
            origin=origin,
            dataset=self.dataset_uri)

        def make_module(package, version, name):
            pipeline_module = database.PipelineModule(
                pipeline=pipeline,
                package=package, version=version, name=name)
            db.add(pipeline_module)
            return pipeline_module

        def make_data_module(name):
            return make_module('data', '0.0', name)

        def make_primitive_module(name):
            if name[0] == '.':
                name = 'd3m.primitives' + name
            return make_module('d3m', '2019.4.4', name)

        def connect(from_module, to_module,
                    from_output='produce', to_input='inputs'):
            db.add(database.PipelineConnection(pipeline=pipeline,
                                               from_module=from_module,
                                               to_module=to_module,
                                               from_output_name=from_output,
                                               to_input_name=to_input))

        input_data = make_data_module('dataset')
        db.add(database.PipelineParameter(
            pipeline=pipeline, module=input_data,
            name='targets', value=pickle.dumps(self.targets),
        ))
        db.add(database.PipelineParameter(
            pipeline=pipeline, module=input_data,
            name='features', value=pickle.dumps(self.features),
        ))

        # FIXME: Denormalize?
        #step0 = make_primitive_module('d3m.primitives.data_transformation.denormalize.Common')
        #connect(input_data, step0, from_output='dataset')

        step1 = make_primitive_module('d3m.primitives.data_transformation.dataset_to_dataframe.Common')
        connect(input_data, step1, from_output='dataset')
        #connect(step0, step1)

        step2 = make_primitive_module('d3m.primitives.data_transformation.column_parser.Common')
        connect(step1, step2)

        step3 = make_primitive_module('d3m.primitives.metalearning.metafeature_extractor.BYU')
        connect(step2, step3)

        db.add(pipeline)
        db.commit()
        logger.info(origin + ' PIPELINE ID: %s', pipeline.id)

        return pipeline.id

    def compute_metafeatures(self, origin):
        db = self.DBSession()
        # Add true and suggested targets
        self._add_target_columns_metadata()
        # Create the metafeatures computing pipeline
        pipeline_id = self._create_metafeatures_pipeline(db, origin)
        # Run the pipeline
        logger.info('Computing Metafeatures')
        metafeature_vector = {x: 0 for x in METAFEATURES_DEFAULT}
        try:
            # TODO Improve the sending of parameters
            outputs = execute(pipeline_id, self.dataset_uri, None, None, None,
                              db_filename=os.path.join(os.environ.get('D3MOUTPUTDIR'), 'temp', 'db.sqlite3'))
                             # TODO: Change this static string path
            for key, value in outputs.items():
                metafeature_results = value.metadata.query(())['data_metafeatures']
                for metafeature_key, metafeature_value in metafeature_results.items():
                    if isinstance(metafeature_value, frozendict.FrozenOrderedDict):
                        for k, v in metafeature_value.items():
                            if 'primitive' not in k:
                                metafeature_name = metafeature_key + '_' + k
                                if metafeature_name in metafeature_vector:
                                    metafeature_vector[metafeature_name] = v
                    else:
                        if metafeature_key in metafeature_vector:
                            metafeature_vector[metafeature_key] = metafeature_value

            return list(metafeature_vector.values())

        except Exception:
            logger.exception('Error running Metafeatures')
            # FIXME: This is a default to address metafeatures not generating features for datasets with numeric targets
            return list(metafeature_vector.values())
        finally:
            db.rollback()
            db.close()



import typing
import logging
import json
import os
from d3m import index
from alphad3m.primitive_loader import load_primitives_hierarchy
from d3m.metadata.hyperparams import Bounded, Enumeration, UniformInt, UniformBool, Uniform, Normal, Union, \
    Constant as ConstantD3M
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.conditions import EqualsCondition, InCondition
from ConfigSpace.hyperparameters import CategoricalHyperparameter, UniformFloatHyperparameter, \
     UniformIntegerHyperparameter, UnParametrizedHyperparameter, Constant, NormalFloatHyperparameter

logger = logging.getLogger(__name__)

HYPERPARAMETERS_FROM_METALEARNING_PATH = os.path.join(os.path.dirname(__file__), '../resource/hyperparams.json')
PRIMITIVES = {}
primitives_info = load_primitives_hierarchy()
for primitive_type in primitives_info:
    for primitive_name in primitives_info[primitive_type]:
        PRIMITIVES[primitive_name] = primitive_type


def is_tunable(primitive_name):
    primitive_type = PRIMITIVES.get(primitive_name, None)

    if primitive_type in {'CLASSIFICATION', 'REGRESSION', 'TIME_SERIES_CLASSIFICATION', 'TIME_SERIES_FORECASTING',
                          'SEMISUPERVISED_CLASSIFICATION', 'COMMUNITY_DETECTION', 'GRAPH_MATCHING', 'LINK_PREDICTION',
                          'VERTEX_CLASSIFICATION', 'OBJECT_DETECTION', 'FEATURE_SELECTION'}:
        return True

    return False


def load_hyperparameters(primitive_name):
    primitive = index.get_primitive(primitive_name)
    hyperparameters_metadata = primitive.metadata.query()['primitive_code']['hyperparams']
    hyperparameter_class = typing.get_type_hints(primitive.__init__)['hyperparams']
    hyperparameters = {}

    if hyperparameter_class:
        for hp_name, hp_value in hyperparameter_class.configuration.items():
            if 'https://metadata.datadrivendiscovery.org/types/TuningParameter' in hyperparameters_metadata[hp_name]['semantic_types']:
                    hyperparameters[hp_name] = hp_value

    return hyperparameters


def load_primitive_configspace(configspace, primitive_name):
    default_configspace = load_default_configspace(primitive_name)
    all_hyperparameters = load_hyperparameters(primitive_name)
    default_hyperparameters = set(default_configspace.get_hyperparameter_names())
    casted_hyperparameters = []
    union_conditions = []
    for hp_name in all_hyperparameters:
        new_hp_name = primitive_name + '|' + hp_name
        if not isinstance(all_hyperparameters[hp_name], Union):
            new_hp = default_configspace.get_hyperparameter(new_hp_name) if new_hp_name in default_hyperparameters \
                else None
            casted_hp = cast_hyperparameter(all_hyperparameters[hp_name], new_hp_name)
            if casted_hp is not None:
                if new_hp is not None and casted_hp.is_legal(new_hp.default_value):
                    casted_hyperparameters.append(new_hp)
                else:
                    casted_hyperparameters.append(casted_hp)
        elif isinstance(all_hyperparameters[hp_name], Union):
            hyperparameter_class = all_hyperparameters[hp_name]
            add_case = False
            case_choices = []
            cases_param_space = []
            for case, case_class in hyperparameter_class.configuration.items():
                new_hp = None
                if new_hp_name + '|' + case in default_hyperparameters:
                    new_hp = default_configspace.get_hyperparameter(new_hp_name + '|' + case)
                casted_hp = cast_hyperparameter(case_class, new_hp_name + '|' + case)
                if casted_hp is not None:
                    add_case = True
                    case_choices.append(case)
                    if new_hp is not None and casted_hp.is_legal(new_hp.default_value):
                        casted_hyperparameters.append(new_hp)
                        cases_param_space.append(new_hp)
                    else:
                        casted_hyperparameters.append(casted_hp)
                        cases_param_space.append(casted_hp)
            if add_case:
                case_hyperparameter = CategoricalHyperparameter(
                    name=new_hp_name + '|case',
                    choices=case_choices,
                    default_value=case_choices[0])
                for i in range(len(case_choices)):
                    union_conditions.append(EqualsCondition(cases_param_space[i], case_hyperparameter, case_choices[i]))
                casted_hyperparameters.append(case_hyperparameter)

    configspace.add_hyperparameters(casted_hyperparameters)

    for condition in default_configspace.get_conditions():
        try:
            configspace.add_condition(condition)
        except Exception as e:
            logger.warning('Not possible to add condition', e)

    for condition in union_conditions:
        try:
            configspace.add_condition(condition)
        except Exception as e:
            logger.warning('Not possible to add condition', e)

    for forbidden in default_configspace.get_forbiddens():
        try:
            configspace.add_forbidden_clause(forbidden)
        except Exception as e:
            logger.warning('Not possible to add forbidden clause', e)


def load_default_configspace(primitive):
    default_config = ConfigurationSpace()

    if primitive in get_hyperparameters_from_metalearnig():
        default_config.add_configuration_space(
            primitive,
            get_configspace_from_metalearning(get_hyperparameters_from_metalearnig()[primitive]),
            '|'
        )

    return default_config


def cast_hyperparameter(hyperparameter, name):
    # From D3M hyperparameters to ConfigSpace hyperparameters
    # TODO: 'Choice' and  'Set' (D3M hyperparameters)
    new_hyperparameter = None

    try:
        if isinstance(hyperparameter, Bounded):
            lower = hyperparameter.lower 
            upper = hyperparameter.upper
            default = hyperparameter.get_default()
            if lower is None:
                lower = default
            if upper is None:
                upper = default * 2 if default > 0 else 10
            if hyperparameter.structural_type == int:
                if not hyperparameter.lower_inclusive:
                    lower += 1
                if not hyperparameter.upper_inclusive:
                    upper -= 1
                new_hyperparameter = UniformIntegerHyperparameter(name, lower, upper, default_value=default)
            else:
                if not hyperparameter.lower_inclusive:
                    lower += 1e-20
                if not hyperparameter.upper_inclusive:
                    upper -= 1e-20
                new_hyperparameter = UniformFloatHyperparameter(name, lower, upper, default_value=default)
        elif isinstance(hyperparameter, UniformBool):
            default = hyperparameter.get_default()
            new_hyperparameter = CategoricalHyperparameter(name, [True, False], default_value=default)
        elif isinstance(hyperparameter, UniformInt):
            lower = hyperparameter.lower
            upper = hyperparameter.upper
            default = hyperparameter.get_default()
            new_hyperparameter = UniformIntegerHyperparameter(name, lower, upper, default_value=default)
        elif isinstance(hyperparameter, Uniform):
            lower = hyperparameter.lower
            upper = hyperparameter.upper
            default = hyperparameter.get_default()
            new_hyperparameter = UniformFloatHyperparameter(name, lower, upper, default_value=default)
        elif isinstance(hyperparameter, Normal):
            default = hyperparameter.get_default()
            new_hyperparameter = NormalFloatHyperparameter(name, default_value=default)
        elif isinstance(hyperparameter, Enumeration):
            values = hyperparameter.values
            default = hyperparameter.get_default()
            new_hyperparameter = CategoricalHyperparameter(name, values, default_value=default)
        elif isinstance(hyperparameter, ConstantD3M):
            default = 'None' if hyperparameter.get_default() is None else hyperparameter.get_default()
            new_hyperparameter = Constant(name, default)
    except Exception as e:
        logger.error(e)

    return new_hyperparameter


def get_hyperparameters_from_metalearnig():
    with open(HYPERPARAMETERS_FROM_METALEARNING_PATH) as fin:
        search_space = json.load(fin)
        return search_space


def get_configspace_from_metalearning(metalearning_entry):
    cs = ConfigurationSpace()
    categorical_and_none_hyperparams = []

    for hyperparam in metalearning_entry:
        if len(metalearning_entry[hyperparam]['choices']) == 1 and metalearning_entry[hyperparam]['default'] is None:
            categorical_and_none_hyperparams.append(
                Constant(hyperparam, 'None')
            )
        else:
            categorical_and_none_hyperparams.append(
                CategoricalHyperparameter(
                    name=hyperparam,
                    choices=metalearning_entry[hyperparam]['choices'],
                    default_value=metalearning_entry[hyperparam]['default'])
            )
    cs.add_hyperparameters(categorical_and_none_hyperparams)

    return cs



"""Unit tests.

Those are supposed to be run without data or primitives available.
"""

import shutil
import tempfile
import unittest
from unittest import mock
from alphad3m.automl import AutoML, Session, TuneHyperparamsJob
from alphad3m.schema import database
from d3m.metadata.problem import PerformanceMetric


class FakePrimitiveBuilder(object):
    def __init__(self, name):
        self._name = name

    def __call__(self):
        return self

    @property
    def metadata(self):
        return self

    def query(self):
        return {
            'name': self._name.rsplit('.', 1)[-1],
            'id': '%s-mocked' % self._name,
            'digest': '00000000',
            'installation': 'mock',
            'description': "This has been mocked out for unit-testing",
            'version': '0.0',
            'python_path': 'mock'
        }


FakePrimitive = FakePrimitiveBuilder('FakePrimitive')


class TestSession(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self._tmp = tempfile.mkdtemp(prefix='d3m_unittest_')
        self._ta2 = AutoML(self._tmp)
        self._problem = {
            'about': {'problemID': 'unittest_problem'},
            'inputs': {
                'data': [
                    {
                        'datasetID': 'unittest_dataset',
                        'targets': [
                            {'resID': '0', 'colName': 'targets'},
                        ],
                    },
                ],
            },
            'problem': {'performance_metrics': [{'metric': PerformanceMetric.F1_MACRO}]}
        }

        db = self._ta2.DBSession()
        self._pipelines = []

        for i in range(5):
            pipeline = database.Pipeline(origin="unittest %d" % i,
                                         dataset='file:///data/test.csv')
            db.add(pipeline)
            mod1 = database.PipelineModule(pipeline=pipeline,
                                           name='tests.tests.FakePrimitive',
                                           package='d3m', version='0.0')
            db.add(mod1)
            mod2 = database.PipelineModule(pipeline=pipeline,
                                           name='tests.tests.FakePrimitive',
                                           package='d3m', version='0.0')
            db.add(mod2)
            db.add(database.PipelineConnection(pipeline=pipeline,
                                               from_module=mod1,
                                               to_module=mod2,
                                               from_output_name='output',
                                               to_input_name='input'))
            self._pipelines.append(pipeline.id)
        db.add(database.Evaluation(
            pipeline_id=self._pipelines[4],
            scores=[
                database.EvaluationScore(metric='F1_MACRO',
                                              value=55.0),
                database.EvaluationScore(metric='EXECUTION_TIME',
                                              value=0.2),
            ],
        ))
        db.commit()
        db.close()

    def tearDown(self):
        shutil.rmtree(self._tmp)

    def test_session(self):
        self._session_test(False)

    def test_session_notuning(self):
        self._session_test(False)

    def _session_test(self, do_tuning):
        db = self._ta2.DBSession()

        def get_job(call):
            assert len(call[-2]) == 1
            assert not call[-1]
            return call[-2][0]

        ta2 = mock.NonCallableMock()
        session = Session(
            ta2,
            self._problem,
            self._ta2.output_folder,
            self._ta2.DBSession
            )
        self._ta2.sessions[session.id] = session

        def compare_scores(ret, expected):
            self.assertEqual(
                [
                    (pipeline.id, score)
                    for (pipeline, score) in ret
                ],
                expected,
            )

        # No pipelines as yet
        self.assertEqual(session.get_top_pipelines(db, PerformanceMetric.F1_MACRO), [])

        # Add scoring pipelines
        session.add_scoring_pipeline(self._pipelines[0])
        session.add_scoring_pipeline(self._pipelines[1])
        session.add_scoring_pipeline(self._pipelines[2])
        session.check_status()
        ta2._run_queue.put.assert_not_called()

        # No pipeline is scored
        compare_scores(session.get_top_pipelines(db, PerformanceMetric.F1_MACRO), [])

        # Pipelines finished scoring
        db.add(database.Evaluation(
            pipeline_id=self._pipelines[0],
            scores=[
                database.EvaluationScore(fold=0,
                                              metric='F1_MACRO',
                                              value=41.5),
                database.EvaluationScore(fold=1,
                                              metric='F1_MACRO',
                                              value=42.5),
                database.EvaluationScore(fold=0,
                                              metric='EXECUTION_TIME',
                                              value=1.3),
                database.EvaluationScore(fold=1,
                                              metric='EXECUTION_TIME',
                                              value=1.5),
            ],
        ))
        db.add(database.Evaluation(
            pipeline_id=self._pipelines[1],
            scores=[
                database.EvaluationScore(fold=0,
                                              metric='F1_MACRO',
                                              value=16.5),
                database.EvaluationScore(fold=1,
                                              metric='F1_MACRO',
                                              value=17.5),
                database.EvaluationScore(fold=0,
                                              metric='EXECUTION_TIME',
                                              value=0.5),
                database.EvaluationScore(fold=1,
                                              metric='EXECUTION_TIME',
                                              value=0.9),
            ],
        ))
        db.commit()

        # Check scores
        compare_scores(session.get_top_pipelines(db, PerformanceMetric.F1_MACRO),
                       [(self._pipelines[0], 42.0),
                        (self._pipelines[1], 17.0)])
        compare_scores(session.get_top_pipelines(db, PerformanceMetric.ACCURACY), [])

        # Finish scoring
        ta2._run_queue.put.assert_not_called()
        session.pipeline_scoring_done(self._pipelines[0])
        ta2._run_queue.put.assert_not_called()
        session.tune_when_ready(3 if do_tuning else 0)
        ta2._run_queue.put.assert_not_called()
        session.pipeline_scoring_done(self._pipelines[1])
        ta2._run_queue.put.assert_not_called()
        session.pipeline_scoring_done(self._pipelines[2])

        # Check tuning jobs were submitted
        if do_tuning:
            ta2._run_queue.put.assert_called()
            self.assertTrue(all(type(get_job(c)) is TuneHyperparamsJob
                                for c in ta2._run_queue.put.mock_calls))
            self.assertEqual(
                [get_job(c).pipeline_id
                 for c in ta2._run_queue.put.mock_calls],
                [self._pipelines[0], self._pipelines[1]]
            )

            # Add tuned pipeline scores
            db.add(database.Evaluation(
                pipeline_id=self._pipelines[2],
                scores=[
                    database.EvaluationScore(fold=0,
                                                  metric='F1_MACRO',
                                                  value=21.0),
                    database.EvaluationScore(fold=1,
                                                  metric='F1_MACRO',
                                                  value=22.0),
                    database.EvaluationScore(fold=0,
                                                  metric='EXECUTION_TIME',
                                                  value=0.9),
                    database.EvaluationScore(fold=1,
                                                  metric='EXECUTION_TIME',
                                                  value=1.1),
                ],
            ))
            db.commit()

            # Signal tuning is done
            ta2._run_queue.put.reset_mock()
            session.pipeline_tuning_done(self._pipelines[0],
                                         self._pipelines[2])
            session.pipeline_tuning_done(self._pipelines[1])
            ta2._run_queue.put.assert_not_called()
        else:
            ta2._run_queue.put.assert_not_called()

        ta2._run_queue.put.assert_not_called()

        # Get top pipelines
        if do_tuning:
            compare_scores(session.get_top_pipelines(db, PerformanceMetric.F1_MACRO),
                           [(self._pipelines[0], 42.0),
                            (self._pipelines[2], 21.5),
                            (self._pipelines[1], 17.0)])
        else:
            compare_scores(session.get_top_pipelines(db, PerformanceMetric.F1_MACRO),
                           [(self._pipelines[0], 42.0),
                            (self._pipelines[1], 17.0)])


if __name__ == '__main__':
    unittest.main()



import logging
import numpy as np
from scipy import stats
from collections import OrderedDict
from alphad3m.primitive_loader import load_primitives_list
from alphad3m.metalearning.database import load_metalearningdb
from alphad3m.metalearning.dataset_miner import get_similar_datasets, get_dataset_id

logger = logging.getLogger(__name__)


IGNORE_PRIMITIVES = {
    # These primitives are static elements in the pipelines, not considered as part of the pattern
    'd3m.primitives.data_transformation.construct_predictions.Common',
    'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
    'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
    'd3m.primitives.data_transformation.denormalize.Common',
    'd3m.primitives.data_transformation.flatten.DataFrameCommon',
    'd3m.primitives.data_transformation.column_parser.Common',
    'd3m.primitives.data_transformation.simple_column_parser.DataFrameCommon',
    'd3m.primitives.data_transformation.do_nothing.DSBOX',
    'd3m.primitives.data_transformation.do_nothing_for_dataset.DSBOX',
    'd3m.primitives.data_transformation.add_semantic_types.Common',
    'd3m.primitives.schema_discovery.profiler.Common',
    'd3m.primitives.schema_discovery.profiler.DSBOX',
    'd3m.primitives.data_cleaning.column_type_profiler.Simon',
    'd3m.primitives.operator.compute_unique_values.Common',
    'd3m.primitives.data_transformation.construct_confidence.Common',
    # We add these primitives internally because they require special connections
    'd3m.primitives.data_transformation.text_reader.Common',
    'd3m.primitives.data_transformation.image_reader.Common'
}


def load_related_pipelines(dataset_path, target_column, task_keywords):
    primitives_by_id = load_primitives_by_id()
    primitives_by_name = load_primitives_by_name()
    all_pipelines = load_metalearningdb()
    similar_datasets = get_similar_datasets('dataprofiles', dataset_path, target_column, task_keywords)
    ignore_primitives_ids = set()
    task_pipelines = []

    for ignore_primitive in IGNORE_PRIMITIVES:
        if ignore_primitive in primitives_by_name:
            ignore_primitives_ids.add(primitives_by_name[ignore_primitive]['id'])

    for pipeline_run in all_pipelines:
        dataset_id_db = get_dataset_id(pipeline_run['problem']['id'])
        if dataset_id_db not in similar_datasets:
            # Skip datasets that are not similar to the target dataset
            continue
        pipeline_primitives = pipeline_run['steps']
        if is_available_primitive(pipeline_primitives, primitives_by_id):
            primitives = filter_primitives(pipeline_primitives, ignore_primitives_ids)
            primitives = [primitives_by_id[p] for p in primitives]  # Use the current names of primitives
            if len(primitives) > 0:
                score = pipeline_run['scores'][0]['normalized']
                metric = pipeline_run['scores'][0]['metric']['metric']
                dataset = pipeline_run['problem']['id']
                task_pipelines.append({'pipeline': primitives, 'score': score, 'metric': metric, 'dataset': dataset,
                                       'pipeline_repr': '_'.join(primitives)})

    logger.info('Found %d related pipelines', len(task_pipelines))

    return task_pipelines


def create_metalearningdb_grammar(task_name, dataset_path, target_column, task_keywords):
    pipelines = load_related_pipelines(dataset_path, target_column, task_keywords)
    patterns, primitives = extract_patterns(pipelines)
    merged_patterns, empty_elements = merge_patterns(patterns)
    grammar = format_grammar(task_name, merged_patterns, empty_elements)

    return grammar, primitives


def format_grammar(task_name, patterns, empty_elements):
    if len(patterns) == 0:
        logger.info('Empty patterns, no grammar have been generated')
        return None

    grammar = 'S -> %s\n' % task_name
    grammar += task_name + ' -> ' + ' | '.join([' '.join(p) for p in patterns])

    for element in sorted(set([e for sublist in patterns for e in sublist])):  # Sort to have a deterministic grammar
        production_rule = element + " -> 'primitive_terminal'"
        if element in empty_elements:
            production_rule += " | 'E'"

        grammar += '\n' + production_rule
    logger.info('Grammar obtained:\n%s', grammar)

    return grammar


def extract_patterns(pipelines, max_nro_patterns=15, min_frequency=3, adtm_threshold=0.5, mean_score_threshold=0.5, ratio_datasets=0.2):
    available_primitives = load_primitives_by_name()
    pipelines = calculate_adtm(pipelines)
    patterns = {}

    for pipeline_data in pipelines:
        if pipeline_data['adtm'] > adtm_threshold:
            # Skip pipelines with average distance to the minimum higher than the threshold
            continue

        primitive_types = [available_primitives[p]['type'] for p in pipeline_data['pipeline']]
        pattern_id = ' '.join(primitive_types)
        if pattern_id not in patterns:
            patterns[pattern_id] = {'structure': primitive_types, 'primitives': set(), 'datasets': set(), 'pipelines': [], 'scores': [], 'adtms': [], 'frequency': 0}
        patterns[pattern_id]['primitives'].update(pipeline_data['pipeline'])
        patterns[pattern_id]['datasets'].add(pipeline_data['dataset'])
        patterns[pattern_id]['pipelines'].append(pipeline_data['pipeline'])
        patterns[pattern_id]['scores'].append(pipeline_data['score'])
        patterns[pattern_id]['adtms'].append(pipeline_data['adtm'])
        patterns[pattern_id]['frequency'] += 1

    logger.info('Found %d different patterns, after creating the portfolio', len(patterns))
    # TODO: Group these removing conditions into a single loop
    # Remove patterns with fewer elements than the minimum frequency
    patterns = {k: v for k, v in patterns.items() if v['frequency'] >= min_frequency}
    logger.info('Found %d different patterns, after removing uncommon patterns', len(patterns))

    # Remove patterns with undesirable primitives (AlphaD3M doesn't have support to handle some of these primitives)
    blacklist_primitives = {'d3m.primitives.data_transformation.dataframe_to_ndarray.Common',
                            'd3m.primitives.data_transformation.list_to_dataframe.DistilListEncoder',
                            'd3m.primitives.data_transformation.ndarray_to_dataframe.Common',
                            'd3m.primitives.data_transformation.horizontal_concat.DSBOX',
                            'd3m.primitives.data_transformation.horizontal_concat.DataFrameCommon',
                            'd3m.primitives.data_transformation.multi_horizontal_concat.Common',
                            'd3m.primitives.data_transformation.conditioner.Conditioner',
                            'd3m.primitives.data_transformation.remove_semantic_types.Common',
                            'd3m.primitives.data_transformation.replace_semantic_types.Common',
                            'd3m.primitives.data_transformation.remove_columns.Common',
                            'd3m.primitives.operator.dataset_map.DataFrameCommon',
                            'd3m.primitives.data_transformation.i_vector_extractor.IVectorExtractor'}
    patterns = {k: v for k, v in patterns.items() if not blacklist_primitives & v['primitives']}
    logger.info('Found %d different patterns, after blacklisting primitives', len(patterns))

    unique_datasets = set()
    for pattern_id in patterns:
        scores = patterns[pattern_id]['scores']
        adtms = patterns[pattern_id]['adtms']
        patterns[pattern_id]['mean_score'] = np.mean(scores)
        patterns[pattern_id]['mean_adtm'] = np.mean(adtms)
        unique_datasets.update(patterns[pattern_id]['datasets'])
    # Remove patterns with low performances
    patterns = {k: v for k, v in patterns.items() if v['mean_score'] >= mean_score_threshold}
    logger.info('Found %d different patterns, after removing low-performance patterns', len(patterns))

    # Remove patterns with low variability
    patterns = {k: v for k, v in patterns.items() if len(v['datasets']) >= len(unique_datasets) * ratio_datasets}
    logger.info('Found %d different patterns, after removing low-variability patterns', len(patterns))

    if len(patterns) > max_nro_patterns:
        logger.info('Found many patterns, selecting top %d (max_nro_patterns)' % max_nro_patterns)
        sorted_patterns = sorted(patterns.items(), key=lambda x: x[1]['mean_score'], reverse=True)
        patterns = {k: v for k, v in sorted_patterns[:max_nro_patterns]}

    primitive_hierarchy = {}
    all_pipelines = []
    all_performances = []
    all_primitives = []
    local_probabilities = {}
    for pattern_id, pattern in patterns.items():
        for primitive in pattern['primitives']:
            primitive_type = available_primitives[primitive]['type']
            if primitive_type not in primitive_hierarchy:
                primitive_hierarchy[primitive_type] = set()
            primitive_hierarchy[primitive_type].add(primitive)
        performances = [1 - x for x in pattern['adtms']]  # Use adtms as performances because their are scaled
        all_pipelines += pattern['pipelines']
        all_primitives += pattern['primitives']
        all_performances += performances
        correlations = calculate_correlations(pattern['primitives'], pattern['pipelines'], performances)
        local_probabilities[pattern_id] = {}
        for primitive, correlation in correlations.items():
            primitive_type = available_primitives[primitive]['type']
            if primitive_type not in local_probabilities[pattern_id]:
                local_probabilities[pattern_id][primitive_type] = {}
            local_probabilities[pattern_id][primitive_type][primitive] = correlation

    correlations = calculate_correlations(set(all_primitives), all_pipelines, all_performances)
    global_probabilities = {}
    for primitive, correlation in correlations.items():
        primitive_type = available_primitives[primitive]['type']
        if primitive_type not in global_probabilities:
            global_probabilities[primitive_type] = {}
        global_probabilities[primitive_type][primitive] = correlation

    # Make deterministic the order of the patterns and hierarchy
    patterns = sorted(patterns.values(), key=lambda x: x['mean_score'], reverse=True)
    primitive_hierarchy = OrderedDict({k: sorted(v) for k, v in sorted(primitive_hierarchy.items(), key=lambda x: x[0])})
    logger.info('Patterns:\n%s', patterns_repr(patterns))
    logger.info('Hierarchy:\n%s', '\n'.join(['%s:\n%s' % (k, ', '.join(v)) for k, v in primitive_hierarchy.items()]))
    patterns = [p['structure'] for p in patterns]
    primitive_probabilities = {'global': global_probabilities, 'local': local_probabilities, 'types': available_primitives}
    primitive_info = {'hierarchy': primitive_hierarchy, 'probabilities': primitive_probabilities}

    return patterns, primitive_info


def calculate_correlations(primitives, pipelines, scores, normalize=True):
    correlations = {}

    for primitive in primitives:
        occurrences = [1 if primitive in pipeline else 0 for pipeline in pipelines]
        correlation_coefficient, p_value = stats.pointbiserialr(occurrences, scores)
        if np.isnan(correlation_coefficient):  # Assign a positive correlation (1) to NaN values
            correlation_coefficient = 1
        if normalize:  # Normalize the Pearson values, from [-1, 1] to [0, 1] range
            correlation_coefficient = (correlation_coefficient - (-1)) / 2  # xi − min(x) / max(x) − min(x)
        correlations[primitive] = round(correlation_coefficient, 4)

    return correlations


def calculate_adtm(pipelines):
    dataset_performaces = {}
    pipeline_performances = {}

    for pipeline_data in pipelines:
        # Even the same dataset can be run under different metrics. So, use the metric to create the id of the dataset
        id_dataset = pipeline_data['dataset'] + '_' + pipeline_data['metric']

        if id_dataset not in dataset_performaces:
            dataset_performaces[id_dataset] = {'min': float('inf'), 'max': float('-inf')}
        performance = pipeline_data['score']

        if performance > dataset_performaces[id_dataset]['max']:
            dataset_performaces[id_dataset]['max'] = performance

        if performance < dataset_performaces[id_dataset]['min']:
            dataset_performaces[id_dataset]['min'] = performance

        id_pipeline = pipeline_data['pipeline_repr']

        if id_pipeline not in pipeline_performances:
            pipeline_performances[id_pipeline] = {}

        if id_dataset not in pipeline_performances[id_pipeline]:
            pipeline_performances[id_pipeline][id_dataset] = pipeline_data['score']
        else:
            # A pipeline can have different performances for a given dataset, choose the best one
            if pipeline_data['score'] > pipeline_performances[id_pipeline][id_dataset]:
                pipeline_performances[id_pipeline][id_dataset] = pipeline_data['score']

    for pipeline_data in pipelines:
        id_pipeline = pipeline_data['pipeline_repr']
        id_dataset_pipeline = pipeline_data['dataset'] + '_' + pipeline_data['metric']
        dtm = 0

        for id_dataset in pipeline_performances[id_pipeline]:  # Iterate over the datasets where the pipeline was used
            minimum = dataset_performaces[id_dataset]['min']
            maximum = dataset_performaces[id_dataset]['max']

            if id_dataset_pipeline == id_dataset:
                score = pipeline_data['score']
            else:
                score = pipeline_performances[id_pipeline][id_dataset]

            if minimum != maximum:
                dtm += (maximum - score) / (maximum - minimum)

        adtm = dtm / len(pipeline_performances[id_pipeline])
        pipeline_data['adtm'] = adtm

    return pipelines


def merge_patterns(grammar_patterns):
    patterns = sorted(grammar_patterns, key=lambda x: len(x), reverse=True)
    empty_elements = set()
    skip_patterns = []

    for pattern in patterns:
        for element in pattern:
            modified_pattern = [e for e in pattern if e != element]
            for current_pattern in patterns:
                if modified_pattern == current_pattern:
                    empty_elements.add(element)
                    skip_patterns.append(modified_pattern)

    for skip_pattern in skip_patterns:
        if skip_pattern in patterns:
            patterns.remove(skip_pattern)

    return patterns, empty_elements


def analyze_distribution(pipelines_metalearningdb):
    available_primitives = load_primitives_by_name()
    primitive_frequency = {}
    primitive_distribution = {}
    logger.info('Analyzing the distribution of primitives')

    for pipeline_data in pipelines_metalearningdb:
        for primitive_name in pipeline_data['pipeline']:
            primitive_type = available_primitives[primitive_name]['type']
            if primitive_type not in primitive_frequency:
                primitive_frequency[primitive_type] = {'primitives': {}, 'total': 0}
            if primitive_name not in primitive_frequency[primitive_type]['primitives']:
                primitive_frequency[primitive_type]['primitives'][primitive_name] = 0
            primitive_frequency[primitive_type]['primitives'][primitive_name] += 1
            primitive_frequency[primitive_type]['total'] += 1

    for primitive_type, primitives_info in primitive_frequency.items():
        if primitive_type not in primitive_distribution:
            primitive_distribution[primitive_type] = []
        for primitive, frequency in sorted(primitives_info['primitives'].items(), key=lambda x: x[1], reverse=True):
            distribution = round(float(frequency) / primitives_info['total'], 4)
            primitive_distribution[primitive_type].append((primitive, distribution))

    logger.info('Distribution:\n%s' % '\n'.join(['%s\n%s' % (k, str(v)) for k, v in primitive_distribution.items()]))
    return primitive_distribution


def is_available_primitive(pipeline_primitives, current_primitives, verbose=False):
    for primitive in pipeline_primitives:
        if primitive['primitive']['id'] not in current_primitives:
            if verbose:
                logger.warning('Primitive %s is not longer available' % primitive['primitive']['python_path'])
            return False
    return True


def filter_primitives(pipeline_steps, ignore_primitives):
    primitives = []

    for pipeline_step in pipeline_steps:
        if pipeline_step['primitive']['id'] not in ignore_primitives:
                primitives.append(pipeline_step['primitive']['id'])

    if len(primitives) > 0 and primitives[0] == '7ddf2fd8-2f7f-4e53-96a7-0d9f5aeecf93':  # Primitive to_numeric.DSBOX
        # This primitive should not be first because it only takes the numeric features, ignoring the remaining ones
        primitives = primitives[1:]

    return primitives


def load_primitives_by_name():
    primitives_by_name = {}
    primitives = load_primitives_list()

    for primitive in primitives:
        primitives_by_name[primitive['python_path']] = {'id': primitive['id'], 'type': primitive['type']}

    return primitives_by_name


def load_primitives_by_id():
    primitives_by_id = {}
    primitives = load_primitives_list()

    for primitive in primitives:
        primitives_by_id[primitive['id']] = primitive['python_path']

    return primitives_by_id


def patterns_repr(patterns):
    patterns_string = []

    for pattern in patterns:
        pretty_string = ''
        pretty_string += 'structure: [%s]' % ', '.join([i for i in pattern['structure']])
        pretty_string += ', frequency: %d' % pattern['frequency']
        if 'mean_score' in pattern:
            pretty_string += ', mean_score: %.3f' % pattern['mean_score']
        if 'mean_adtm' in pattern:
            pretty_string += ', mean_adtm: %.3f' % pattern['mean_adtm']
        patterns_string.append(pretty_string)

    return '\n'.join(patterns_string)


def test_dataset(dataset_id, task_name='TASK'):
    from os.path import join
    import json
    dataset_folder_path = join('/Users/rlopez/D3M/datasets/seed_datasets_current/', dataset_id)
    dataset_path = join(dataset_folder_path, 'TRAIN/dataset_TRAIN/tables/learningData.csv')
    problem_path = join(dataset_folder_path, 'TRAIN/problem_TRAIN/problemDoc.json')

    with open(problem_path) as fin:
        problem_doc = json.load(fin)
        task_keywords = problem_doc['about']['taskKeywords']
        target_column = problem_doc['inputs']['data'][0]['targets'][0]['colName']
    logger.info('Evaluating dataset %s with task keywords=%s' % (dataset_id, str(task_keywords)))
    create_metalearningdb_grammar(task_name, dataset_path, target_column, task_keywords)
    #analyze_distribution(load_related_pipelines(dataset_path, target_column, task_keywordsn))


if __name__ == '__main__':
    test_dataset('185_baseball_MIN_METADATA')



import os
import logging
import pickle
import d3m.runtime
import d3m.metadata.base
from os.path import join
from sqlalchemy.orm import joinedload
from d3m.container import Dataset, DataFrame
from d3m.metadata import base as metadata_base
from alphad3m.schema import database, convert


logger = logging.getLogger(__name__)


@database.with_db
def train(pipeline_id, dataset, problem, storage_dir, steps_to_expose, msg_queue, db):
    # Get pipeline from database
    pipeline = (
        db.query(database.Pipeline)
            .filter(database.Pipeline.id == pipeline_id)
            .options(joinedload(database.Pipeline.modules),
                     joinedload(database.Pipeline.connections))
    ).one()

    logger.info('About to train pipeline, id=%s, dataset=%r',
                pipeline_id, dataset)

    # Load data
    dataset = Dataset.load(dataset)
    logger.info('Loaded dataset')

    # Training step - fit pipeline on training data
    logger.info('Running training')

    d3m_pipeline = d3m.metadata.pipeline.Pipeline.from_json_structure(
        convert.to_d3m_json(pipeline),
    )

    expose_outputs = True if len(steps_to_expose) > 0 else False

    fitted_pipeline, predictions, results = d3m.runtime.fit(d3m_pipeline, [dataset], problem_description=problem,
                                                            context=metadata_base.Context.TESTING,
                                                            volumes_dir=os.environ.get('D3MSTATICDIR', None),
                                                            random_seed=0,
                                                            expose_produced_outputs=expose_outputs)

    results.check_success()

    logger.info('Storing fit results at %s', storage_dir)
    for step_id in results.values:
        if step_id in steps_to_expose and isinstance(results.values[step_id], DataFrame):
            results.values[step_id].to_csv(join(storage_dir, 'fit_%s_%s.csv' % (pipeline_id, step_id)))

    with open(join(storage_dir, 'fitted_solution_%s.pkl' % pipeline_id), 'wb') as fout:
        pickle.dump(fitted_pipeline, fout)



'''
Board class.
Board data:
  1=x, -1=o, 0=empty
  first dim is column , 2nd is row:
     pieces[1][2] is the square in column 2, row 3.
Squares are stored and manipulated as (x,y) tuples.
x is the column, y is the row.
'''

import numpy as np
import logging

logger = logging.getLogger(__name__)


class Board():

    def __init__(self, m=30, grammar={}, pipeline_size=6, metric='f1macro', win_threshold=0.6):
        "Set up initial board configuration."

        self.terminals = grammar['TERMINALS']
        self.non_terminals = grammar['NON_TERMINALS']
        self.start = grammar['START']
        self.m = m  # Number of metafeatures
        self.p = pipeline_size #max length of pipeline
        self.valid_moves = [i for i, j in sorted(grammar['RULES'].items(), key=lambda x: x[1])]
        self.rules = grammar['RULES']
        self.rules_lookup = grammar['RULES_LOOKUP']
        #logger.info('NUMBER of VALID MOVES %s', len(self.valid_moves))
        
        # Create the empty board array.
        self.pieces_m = [0] * self.m
        start_pipeline = [self.terminals[p] if p in self.terminals else self.non_terminals[p] for p in
                    [self.start]]

        self.pieces_p = [0] * (self.p - len(start_pipeline)) + start_pipeline

        if 'error' in metric.lower():  # TODO: change it
            win_threshold = -1 * win_threshold
        self.win_threshold = win_threshold

    # add [][] indexer syntax to the Board
    def __getitem__(self, index):
        return self.pieces_p[index]

    def get_pipeline(self, board):
        return board[self.m:]

    def get_metafeatures(self, board):
        return board[0:self.m]

    def set_metafeatures(self, board):
        self.pieces_m = board[0:self.m]

    def set_pipeline(self, board):
        self.pieces_p = board[self.m:]

    def is_terminal_pipeline(self):
        for p in self.pieces_p:
            if p != 0 and p in list(self.non_terminals.values()):  # Empty symbol ID = 0
                return False
        return True
    
    def findWin(self, player, eval_val=None):
        """Find win of the given color in row, column, or diagonal
        (1 for x, -1 for o)"""
        if not any(self[0:]):
            return False
        if eval_val == float('inf'):
            return False

        return eval_val >= self.win_threshold
    
    def get_legal_moves(self):
        """Returns all the legal moves.
        """

        valid_moves = np.asarray([0]*len(self.valid_moves))
        pipeline = [p for p in self.pieces_p if p != 0]

        for p in pipeline:
            if p in list(self.non_terminals.values()):
                #logger.info('GET LEGAL MOVES %s', p)
                rules = [self.rules[key]-1 for key in self.rules_lookup[list(self.non_terminals.keys())[p-1]] if len(self.next_state(self.rules[key]-1)) <= self.p and len(set(self.next_state(self.rules[key]-1))) == len(self.next_state(self.rules[key]-1))]
                np.put(valid_moves, rules, [1]*len(rules))

        return valid_moves.tolist()
        
    def has_legal_moves(self):
        return len(np.where(np.asarray(self.get_legal_moves()) == 1)[0]) > 0

    def next_state(self, action):
        s = self.valid_moves[action]
        nt = self.non_terminals[s[:s.index('-')].strip()]
        r = [self.non_terminals[p] if p in self.non_terminals.keys() else self.terminals[p] for p in s[s.index('-')+2:].strip().split(' ')]
        r = [x for x in r if x != 0]
        s = []
        for p in self.pieces_p:
            if p == 0:
                continue
            
            if p == nt:
                s += r
            else:
                s.append(p)

        return s

    def get_pipeline_primitives(self, pipeline):
        #logger.info('PIPELINE PRIMITIVES FOR %s', pipeline)
        return [list(self.terminals.keys())[list(self.terminals.values()).index(i)] if i in list(self.terminals.values()) else list(self.non_terminals.keys())[list(self.non_terminals.values()).index(i)] for i in pipeline if not i == 0]

    def get_train_board(self):
        logger.info('TRAIN BOARD: %s', '|'.join(self.get_pipeline_primitives(self.pieces_p)))
        pipeline = [0]*(len(self.terminals)+len(self.non_terminals))

        for p in self.pieces_p:
            if p != 0:
                 pipeline[p] = 1
            #pipeline[p-1] = 1

        return self.pieces_m + pipeline

    def get_board_size(self):
        return self.m+(len(self.terminals)+len(self.non_terminals))
    
    def execute_move(self, action, player):
        """Perform the given move on the board;
        color gives the color of the piece to play (1=x,-1=o)
        """
        logger.info('MOVE ACTION: %s', self.valid_moves[action])
        s = self.next_state(action)
        
        s = [0] * (self.p - len(s)) + s

        self.pieces_p = s





import logging
import os
import pickle
from os.path import join
from d3m.container import Dataset, DataFrame
from alphad3m.schema import database

logger = logging.getLogger(__name__)


@database.with_db
def test(pipeline_id, dataset, storage_dir, steps_to_expose, msg_queue, db):
    dataset = Dataset.load(dataset)
    logger.info('Loaded dataset')

    runtime = None
    with open(os.path.join(storage_dir, 'fitted_solution_%s.pkl' % pipeline_id), 'rb') as fin:
        runtime = pickle.load(fin)

    results = runtime.produce(inputs=[dataset], outputs_to_expose=steps_to_expose)
    results.check_success()

    logger.info('Storing produce results at %s', storage_dir)
    for step_id in results.values:
        if step_id in steps_to_expose and isinstance(results.values[step_id], DataFrame):
            results.values[step_id].to_csv(join(storage_dir, 'produce_%s_%s.csv' % (pipeline_id, step_id)))



import os
import pickle
import logging
from alphad3m.schema import database
from d3m import index
from d3m.container import Dataset, DataFrame, ndarray, List
from alphad3m.utils import is_collection, get_collection_type


# Use a headless matplotlib backend
os.environ['MPLBACKEND'] = 'Agg'
logger = logging.getLogger(__name__)

CONTAINER_CAST = {
    Dataset: {
        DataFrame: 'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
        ndarray: ('d3m.primitives.data_transformation.dataset_to_dataframe.Common'
                  '|d3m.primitives.data_transformation.dataframe_to_ndarray.Common'),
        List: ('d3m.primitives.data_transformation.dataset_to_dataframe.Common'
               '|d3m.primitives.data_transformation.dataframe_to_list.Common')
    },
    DataFrame: {
        Dataset: '',
        ndarray: 'd3m.primitives.data_transformation.dataframe_to_ndarray.Common',
        List: 'd3m.primitives.data_transformation.dataframe_to_list.Common'
    },
    ndarray: {
        Dataset: '',
        DataFrame: 'd3m.primitives.data_transformation.ndarray_to_dataframe.Common',
        List: 'd3m.primitives.data_transformation.ndarray_to_list.Common'
    },
    List: {
        Dataset: '',
        DataFrame: 'd3m.primitives.data_transformation.list_to_dataframe.Common',
        ndarray: 'd3m.primitives.data_transformation.list_to_ndarray.Common',
    }
}


def make_pipeline_module(db, pipeline, name, package='d3m', version='2019.10.10'):
    pipeline_module = database.PipelineModule(pipeline=pipeline, package=package, version=version, name=name)
    db.add(pipeline_module)
    return pipeline_module


def make_data_module(db, pipeline, targets, features):
    input_data = make_pipeline_module(db, pipeline, 'dataset', 'data', '0.0')
    db.add(database.PipelineParameter(
        pipeline=pipeline, module=input_data,
        name='targets', value=pickle.dumps(targets),
    ))
    db.add(database.PipelineParameter(
        pipeline=pipeline, module=input_data,
        name='features', value=pickle.dumps(features),
    ))
    return input_data


def connect(db, pipeline, from_module, to_module, from_output='produce', to_input='inputs'):
    if 'index' not in from_output:
        if not from_module.name.startswith('dataset'):
            from_module_prim = index.get_primitive(from_module.name)
            from_module_output = from_module_prim.metadata.query()['primitive_code']['class_type_arguments']['Outputs']
        else:
            from_module_output = Dataset

        to_module_prim = index.get_primitive(to_module.name)
        to_module_input = to_module_prim.metadata.query()['primitive_code']['class_type_arguments']['Inputs']
        arguments = to_module_prim.metadata.query()['primitive_code']['arguments']

        if to_input not in arguments:
             raise NameError('Argument %s not found in %s' % (to_input, to_module.name))

        if to_module.name == 'd3m.primitives.feature_extraction.audio_transfer.DistilAudioTransfer':
            from_output = 'produce_collection'

        if from_module_output != to_module_input and \
                from_module.name != 'd3m.primitives.data_transformation.audio_reader.DistilAudioDatasetLoader':
            # FIXME: DistilAudioDatasetLoader has a bug https://github.com/uncharted-distil/distil-primitives/issues/294
            cast_module_steps = CONTAINER_CAST[from_module_output][to_module_input]
            if cast_module_steps:
                for cast_step in cast_module_steps.split('|'):
                    cast_module = make_pipeline_module(db, pipeline, cast_step)
                    db.add(database.PipelineConnection(pipeline=pipeline,
                                                       from_module=from_module,
                                                       to_module=cast_module,
                                                       from_output_name=from_output,
                                                       to_input_name='inputs'))
                    from_module = cast_module
            else:
                raise TypeError('Incompatible connection types: %s and %s for primitives %s and %s' % (
                    str(from_module_output), str(to_module_input), from_module.name, to_module.name))

    db.add(database.PipelineConnection(pipeline=pipeline,
                                       from_module=from_module,
                                       to_module=to_module,
                                       from_output_name=from_output,
                                       to_input_name=to_input))


def set_hyperparams(db, pipeline, module, **hyperparams):
    db.add(database.PipelineParameter(
        pipeline=pipeline, module=module,
        name='hyperparams', value=pickle.dumps(hyperparams),
    ))


def change_default_hyperparams(db, pipeline, primitive_name, primitive, learner_index=None):
    if primitive_name == 'd3m.primitives.data_transformation.one_hot_encoder.SKlearn':
        set_hyperparams(db, pipeline, primitive, use_semantic_types=True, return_result='replace', handle_unknown='ignore')
    elif primitive_name == 'd3m.primitives.data_cleaning.imputer.SKlearn':
        set_hyperparams(db, pipeline, primitive, use_semantic_types=True, return_result='replace', strategy='most_frequent', error_on_no_input=False)
    elif primitive_name.endswith('.SKlearn') and not (primitive_name.startswith('d3m.primitives.classification.') or
                                                      primitive_name.startswith('d3m.primitives.regression.')):
        set_hyperparams(db, pipeline, primitive, use_semantic_types=True, return_result='replace')
    elif primitive_name == 'd3m.primitives.data_transformation.enrich_dates.DistilEnrichDates':
        set_hyperparams(db, pipeline, primitive, replace=True)
    elif primitive_name == 'd3m.primitives.data_transformation.encoder.DSBOX':
        set_hyperparams(db, pipeline, primitive, n_limit=50)
    elif primitive_name == 'd3m.primitives.data_transformation.splitter.DSBOX':
        set_hyperparams(db, pipeline, primitive, threshold_row_length=2000)
    elif primitive_name == 'd3m.primitives.data_transformation.encoder.DistilTextEncoder':
        set_hyperparams(db, pipeline, primitive, encoder_type='tfidf')
    elif primitive_name == 'd3m.primitives.classification.text_classifier.DistilTextClassifier':
        set_hyperparams(db, pipeline, primitive, metric='accuracy')
    elif primitive_name == 'd3m.primitives.data_transformation.satellite_image_loader.DistilSatelliteImageLoader':
        set_hyperparams(db, pipeline, primitive, return_result='replace')
    elif primitive_name == 'd3m.primitives.clustering.k_means.DistilKMeans':
        set_hyperparams(db, pipeline, primitive, cluster_col_name='Class')
    elif primitive_name == 'd3m.primitives.data_transformation.adjacency_spectral_embedding.JHU':
        set_hyperparams(db, pipeline, primitive, use_attributes=True, max_dimension=5)
    elif primitive_name == 'd3m.primitives.graph_clustering.gaussian_clustering.JHU':
        set_hyperparams(db, pipeline, primitive, max_clusters=10)
    elif primitive_name == 'd3m.primitives.time_series_classification.convolutional_neural_net.LSTM_FCN':
        set_hyperparams(db, pipeline, primitive, epochs=1)
    elif primitive_name == 'd3m.primitives.time_series_forecasting.nbeats.DeepNeuralNetwork':
        set_hyperparams(db, pipeline, primitive, window_sampling_limit_multiplier=200, batch_size=10)
    elif primitive_name == 'd3m.primitives.time_series_forecasting.esrnn.RNN':
        set_hyperparams(db, pipeline, primitive, auto_tune=True)
    elif primitive_name == 'd3m.primitives.semisupervised_classification.iterative_labeling.AutonBox':
        if learner_index is not None:
            set_hyperparams(db, pipeline, primitive,  blackbox={'type': 'PRIMITIVE', 'data': learner_index})


def need_entire_dataframe(primitives):
    for primitive in primitives:
        if primitive in {'d3m.primitives.data_transformation.time_series_to_list.DSBOX',
                         'd3m.primitives.feature_extraction.random_projection_timeseries_featurization.DSBOX',
                         'd3m.primitives.data_transformation.dataframe_to_tensor.DSBOX',
                         'd3m.primitives.feature_extraction.resnet50_image_feature.DSBOX'}:
            return True
    return False


def need_parsed_targets(primitives):
    for primitive in primitives:
        if primitive in {'d3m.primitives.collaborative_filtering.link_prediction.DistilCollaborativeFiltering',
                         'd3m.primitives.time_series_forecasting.esrnn.RNN',
                         'd3m.primitives.time_series_forecasting.nbeats.DeepNeuralNetwork',
                         'd3m.primitives.time_series_forecasting.vector_autoregression.VAR'}:
            return True
    return False


def is_learner(primitive):
    if primitive.startswith('d3m.primitives.classification.'):
        return True
    return False


def is_linear_pipeline(primitives):
    for primitive in primitives:
        if primitive in {'d3m.primitives.data_transformation.load_single_graph.DistilSingleGraphLoader',
                         'd3m.primitives.data_transformation.load_graphs.DistilGraphLoader',
                         'd3m.primitives.data_transformation.load_graphs.JHU',
                         'd3m.primitives.data_transformation.load_edgelist.DistilEdgeListLoader',
                         'd3m.primitives.graph_matching.seeded_graph_matching.JHU'}:
            return True

    return False


def is_template_pipeline(primitives):
    for primitive in primitives:
        if primitive in {'d3m.primitives.graph_matching.euclidean_nomination.JHU',
                         'd3m.primitives.feature_extraction.yolo.DSBOX',
                         'd3m.primitives.object_detection.retina_net.ObjectDetectionRN'}:
            return True

    return False


def process_template(db, input_data, pipeline, pipeline_template, count_template_steps=0, prev_step=None):
    prev_steps = {}
    for pipeline_step in pipeline_template['steps']:
        if pipeline_step['type'] == 'PRIMITIVE':
            step = make_pipeline_module(db, pipeline, pipeline_step['primitive']['python_path'])
            if 'outputs' in pipeline_step:
                for output in pipeline_step['outputs']:
                    prev_steps['steps.%d.%s' % (count_template_steps, output['id'])] = step

            count_template_steps += 1
            if 'hyperparams' in pipeline_step:
                hyperparams = {}
                for hyper, desc in pipeline_step['hyperparams'].items():
                    hyperparams[hyper] = {'type': desc['type'], 'data': desc['data']}
                set_hyperparams(db, pipeline, step, **hyperparams)
        else:
            # TODO In the future we should be able to handle subpipelines
            break
        if prev_step:
            if 'arguments' in pipeline_step:
                for argument, desc in pipeline_step['arguments'].items():
                    connect(db, pipeline, prev_steps[desc['data']], step,
                            from_output=desc['data'].split('.')[-1], to_input=argument)
            # index is a special connection to keep the order of steps in fixed pipeline templates
            connect(db, pipeline, prev_step, step, from_output='index', to_input='index')
        else:
            connect(db, pipeline, input_data, step, from_output='dataset')
        prev_step = step

    return prev_step, count_template_steps


def add_semantic_types(db, metadata, pipeline, pipeline_template, prev_step):
    count_steps = 0
    if pipeline_template is None:
        for semantic_type, columns in metadata['semantictypes_indices'].items():
            step_add_type = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.'
                                                               'add_semantic_types.Common')
            count_steps += 1
            set_hyperparams(db, pipeline, step_add_type, columns=columns, semantic_types=[semantic_type])
            connect(db, pipeline, prev_step, step_add_type)
            prev_step = step_add_type
    else:
        step_add_type = make_pipeline_module(db, pipeline, 'd3m.primitives.schema_discovery.profiler.Common')
        count_steps += 1
        connect(db, pipeline, prev_step, step_add_type)
        prev_step = step_add_type
    return prev_step, count_steps


def add_file_readers(db, pipeline, prev_step, dataset_path):
    last_step = prev_step
    count_steps = 0
    if get_collection_type(dataset_path) == 'text':
        text_reader = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.text_reader.Common')
        count_steps += 1
        set_hyperparams(db, pipeline, text_reader, return_result='replace')
        connect(db, pipeline, prev_step, text_reader)
        last_step = text_reader

    elif get_collection_type(dataset_path) == 'image':
        image_reader = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.image_reader.Common')
        count_steps += 1
        set_hyperparams(db, pipeline, image_reader, return_result='replace')
        connect(db, pipeline, prev_step, image_reader)
        last_step = image_reader

    return last_step, count_steps


def add_rocauc_primitives(pipeline, current_step, prev_step, target_step, dataframe_step, learner_index, db):
    horizontal_concat = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.horizontal_concat.DataFrameCommon')
    # 'index'  is an artificial connection, just to guarantee the order of the steps
    connect(db, pipeline, current_step, horizontal_concat, from_output='index', to_input='index')
    connect(db, pipeline, prev_step, horizontal_concat, to_input='left')
    connect(db, pipeline, target_step, horizontal_concat, to_input='right')

    compute_values = make_pipeline_module(db, pipeline, 'd3m.primitives.operator.compute_unique_values.Common')
    connect(db, pipeline, horizontal_concat, compute_values)

    construct_confidence = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.construct_confidence.Common')
    set_hyperparams(db, pipeline, construct_confidence, primitive_learner={"type": "PRIMITIVE", "data": learner_index})
    connect(db, pipeline, compute_values, construct_confidence)
    connect(db, pipeline, dataframe_step, construct_confidence, to_input='reference')


def add_previous_primitives(db, pipeline, primitives, prev_step):
    remaining_primitives = []
    count_steps = 0
    for primitive in primitives:
        if need_entire_dataframe([primitive]):
            step_add_type = make_pipeline_module(db, pipeline, primitive)
            count_steps += 1
            connect(db, pipeline, prev_step, step_add_type)
            prev_step = step_add_type
        else:
            remaining_primitives.append(primitive)

    return prev_step, remaining_primitives, count_steps


def select_parsed_semantic_types(primitives, pipeline, step, db):
    for primitive in primitives:
        if primitive in {'d3m.primitives.data_transformation.grouping_field_compose.Common'}:
            set_hyperparams(db, pipeline, step,
                            parse_semantic_types=['http://schema.org/Boolean', 'http://schema.org/Integer',
                                                  'http://schema.org/Float', 'http://schema.org/DateTime',
                                                  'https://metadata.datadrivendiscovery.org/types/FloatVector']
                            )
        elif primitive in {'d3m.primitives.time_series_forecasting.arima.DSBOX'}:
            set_hyperparams(db, pipeline, step,
                            parse_semantic_types=['http://schema.org/Boolean', 'http://schema.org/Integer',
                                                  'http://schema.org/Float']
                            )


class BaseBuilder:

    def make_d3mpipeline(self, primitives, origin, dataset, pipeline_template, targets, features,
                         metadata, metrics=[], DBSession=None):
        # TODO: Remove parameters 'features' and 'targets' are not needed
        db = DBSession()
        dataset_path = dataset[7:]
        origin_name = '%s (%s)' % (origin, ', '.join([p.replace('d3m.primitives.', '') for p in primitives]))
        pipeline = database.Pipeline(origin=origin_name, dataset=dataset)
        count_steps = 0
        learner_index = None

        if is_linear_pipeline(primitives):
            return self.make_linear_pipeline(primitives, dataset, targets, features, origin, db)

        if is_template_pipeline(primitives):
            return self.make_template_pipeline(primitives, dataset, targets, features, origin, db)

        try:
            input_data = make_data_module(db, pipeline, targets, features)

            step0 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.denormalize.Common')

            if metadata['large_rows']:
                step_sample = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.dataset_sample.Common')
                set_hyperparams(db, pipeline, step_sample, sample_size=metadata['sample_size'])
                connect(db, pipeline, step0, step_sample)
                count_steps += 1
                step0 = step_sample

            if not pipeline_template:
                connect(db, pipeline, input_data, step0, from_output='dataset')
            else:
                template_step, template_count = process_template(db, input_data, pipeline, pipeline_template)
                connect(db, pipeline, template_step, step0)
                count_steps += template_count

            step1 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.dataset_to_dataframe.Common')
            connect(db, pipeline, step0, step1)
            count_steps += 1
            prev_step = step1

            if is_collection(dataset_path) and not need_entire_dataframe(primitives):
                prev_step, reader_steps = add_file_readers(db, pipeline, prev_step, dataset_path)
                count_steps += reader_steps

            if len(metadata['semantictypes_indices']) > 0:
                prev_step, semantic_steps = add_semantic_types(db, metadata, pipeline, pipeline_template, prev_step)
                count_steps += semantic_steps

            dataframe_step = prev_step
            if need_entire_dataframe(primitives):
                prev_step, primitives, primitive_steps = add_previous_primitives(db, pipeline, primitives, prev_step)
                count_steps += primitive_steps

            if metadata['large_columns']:
                # TODO: Remove this when https://gitlab.com/datadrivendiscovery/common-primitives/-/issues/149 is fixed
                step2 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.column_parser.DistilColumnParser')
            else:
                step2 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.column_parser.Common')
                select_parsed_semantic_types(primitives, pipeline, step2, db)

            connect(db, pipeline, prev_step, step2)
            count_steps += 1

            step3 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common')
            set_hyperparams(db, pipeline, step3,
                            semantic_types=['https://metadata.datadrivendiscovery.org/types/Attribute'],
                            exclude_columns=metadata['exclude_columns'])
            connect(db, pipeline, step2, step3)
            count_steps += 1

            step4 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common')
            set_hyperparams(db, pipeline, step4, semantic_types=['https://metadata.datadrivendiscovery.org/types/TrueTarget'])
            if need_parsed_targets(primitives):
                connect(db, pipeline, step2, step4)
            else:
                connect(db, pipeline, dataframe_step, step4)
            count_steps += 1

            current_step = prev_step = preprev_step = step3
            for primitive in primitives:
                current_step = make_pipeline_module(db, pipeline, primitive)
                change_default_hyperparams(db, pipeline, primitive, current_step, learner_index)

                if 'semisupervised_classification' in primitive:
                    connect(db, pipeline, preprev_step, current_step)
                    connect(db, pipeline, prev_step, current_step, from_output='index', to_input='index')
                else:
                    connect(db, pipeline, prev_step, current_step)

                if 'outputs' in index.get_primitive(primitive).metadata.query()['primitive_code']['arguments']:
                    connect(db, pipeline, step4, current_step, to_input='outputs')

                preprev_step = prev_step
                prev_step = current_step
                count_steps += 1
                if is_learner(primitive):
                    learner_index = count_steps

            if 'ROC_AUC' in metrics[0]['metric'].name:
                add_rocauc_primitives(pipeline, current_step, preprev_step, step4, dataframe_step, learner_index, db)
            else:
                step5 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.construct_predictions.Common')
                connect(db, pipeline, current_step, step5)
                connect(db, pipeline, dataframe_step, step5, to_input='reference')

            db.add(pipeline)
            db.commit()
            logger.info('%s PIPELINE ID: %s', origin, pipeline.id)
            return pipeline.id
        except Exception as e:
            logger.exception('Error creating pipeline id=%s, primitives=%s', pipeline.id, str(primitives), exc_info=e)
            return None
        finally:
            db.close()

    def make_linear_pipeline(self, primitives, dataset, targets, features, origin, db):
        origin_name = '%s (%s)' % (origin, ', '.join([p.replace('d3m.primitives.', '') for p in primitives]))
        pipeline = database.Pipeline(origin=origin_name, dataset=dataset)

        try:
            input_data = make_data_module(db, pipeline, targets, features)
            step0 = make_pipeline_module(db, pipeline, primitives[0])
            connect(db, pipeline, input_data, step0, from_output='dataset')
            primitives = primitives[1:]
            prev_step = step0

            for primitive in primitives:
                current_step = make_pipeline_module(db, pipeline, primitive)
                change_default_hyperparams(db, pipeline, primitive, current_step)
                connect(db, pipeline, prev_step, current_step)
                if 'outputs' in index.get_primitive(primitive).metadata.query()['primitive_code']['arguments']:
                    connect(db, pipeline, prev_step, current_step, to_input='outputs', from_output='produce_target')
                prev_step = current_step

            db.add(pipeline)
            db.commit()
            logger.info('%s PIPELINE ID: %s', origin, pipeline.id)
            return pipeline.id

        except:
            logger.exception('Error creating pipeline id=%s, primitives=%s', pipeline.id, str(primitives))
            return None
        finally:
            db.close()

    def make_template_pipeline(self, primitives, dataset, targets, features, origin, db):
        # There are some primitives hard to be connected. So, use them in a pipeline template
        origin_name = '%s (%s)' % (origin, ', '.join([p.replace('d3m.primitives.', '') for p in primitives]))
        pipeline = database.Pipeline(origin=origin_name, dataset=dataset)

        try:
            input_data = make_data_module(db, pipeline, targets, features)
            if primitives[0] == 'd3m.primitives.graph_matching.euclidean_nomination.JHU':
                step0 = make_pipeline_module(db, pipeline,
                                             'd3m.primitives.data_transformation.dataset_to_dataframe.Common')
                connect(db, pipeline, input_data, step0)

                step1 = make_pipeline_module(db, pipeline,
                                             'd3m.primitives.data_transformation.dataset_to_dataframe.Common')
                set_hyperparams(db, pipeline, step1, dataframe_resource='1')
                connect(db, pipeline, input_data, step1)

                step2 = make_pipeline_module(db, pipeline,
                                             'd3m.primitives.data_transformation.dataset_to_dataframe.Common')
                set_hyperparams(db, pipeline, step2, dataframe_resource='2')
                connect(db, pipeline, input_data, step2)

                step3 = make_pipeline_module(db, pipeline, primitives[0])
                connect(db, pipeline, step1, step3, to_input='inputs_1')
                connect(db, pipeline, step2, step3, to_input='inputs_2')
                connect(db, pipeline, step0, step3, to_input='reference')

            elif primitives[0] == 'd3m.primitives.feature_extraction.yolo.DSBOX':
                # FIXME: It's simple to generalize and use in make_d3mpipeline
                step0 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.denormalize.Common')
                connect(db, pipeline, input_data, step0, from_output='dataset')
                step1 = make_pipeline_module(db, pipeline,
                                             'd3m.primitives.data_transformation.dataset_to_dataframe.Common')
                connect(db, pipeline, step0, step1)

                step2 = make_pipeline_module(db, pipeline,
                                             'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common')
                set_hyperparams(db, pipeline, step2,
                                semantic_types=['https://metadata.datadrivendiscovery.org/types/PrimaryMultiKey',
                                                'https://metadata.datadrivendiscovery.org/types/FileName']
                                )
                connect(db, pipeline, step1, step2)

                step3 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.'
                                                           'extract_columns_by_semantic_types.Common')
                set_hyperparams(db, pipeline, step3,
                                semantic_types=['https://metadata.datadrivendiscovery.org/types/TrueTarget'],
                                )
                connect(db, pipeline, step1, step3)

                step4 = make_pipeline_module(db, pipeline, primitives[0])
                connect(db, pipeline, step2, step4)
                connect(db, pipeline, step3, step4, to_input='outputs')

                step5 = make_pipeline_module(db, pipeline,
                                             'd3m.primitives.data_transformation.construct_predictions.Common')
                connect(db, pipeline, step4, step5)
                connect(db, pipeline, step2, step5, to_input='reference')

            elif primitives[0] == 'd3m.primitives.object_detection.retina_net.ObjectDetectionRN':
                # FIXME: It's simple to generalize and use in make_linear_pipeline
                step0 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.denormalize.Common')
                connect(db, pipeline, input_data, step0, from_output='dataset')
                step1 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.'
                                                           'dataset_to_dataframe.Common')
                set_hyperparams(db, pipeline, step1)
                connect(db, pipeline, step0, step1)

                step2 = make_pipeline_module(db, pipeline, primitives[0])
                connect(db, pipeline, step1, step2)
                connect(db, pipeline, step1, step2, to_input='outputs')

            db.add(pipeline)
            db.commit()
            logger.info('%s PIPELINE ID: %s', origin, pipeline.id)
            return pipeline.id

        except:
            logger.exception('Error creating pipeline id=%s, primitives=%s', pipeline.id, str(primitives))
            return None
        finally:
            db.close()


class AudioBuilder(BaseBuilder):

    def make_d3mpipeline(self, primitives, origin, dataset, pipeline_template, targets, features,
                         metadata, metrics=[], DBSession=None):
        db = DBSession()
        origin_name = '%s (%s)' % (origin, ', '.join([p.replace('d3m.primitives.', '') for p in primitives]))
        pipeline = database.Pipeline(origin=origin_name, dataset=dataset)
        count_steps = 0

        try:
            input_data = make_data_module(db, pipeline, targets, features)
            prev_step = input_data
            while True:
                # Use primitives that need Dataset object as inputs
                if index.get_primitive(primitives[0]).metadata.query()['primitive_code']['arguments']['inputs']['type'] \
                        != Dataset:
                    break
                current_step = make_pipeline_module(db, pipeline, primitives[0])
                connect(db, pipeline, prev_step, current_step)
                prev_step = current_step
                count_steps += 1
                primitives = primitives[1:]

            dataframe_step = current_step
            step1 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.column_parser.Common')
            set_hyperparams(db, pipeline, step1, parse_semantic_types=[
                        "http://schema.org/Boolean",
                        "http://schema.org/Integer",
                        "http://schema.org/Float",
                        "https://metadata.datadrivendiscovery.org/types/FloatVector"
                    ]
            )
            connect(db, pipeline, prev_step, step1)
            count_steps += 1

            step2 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common')
            set_hyperparams(db, pipeline, step2, semantic_types=['https://metadata.datadrivendiscovery.org/types/TrueTarget'])
            connect(db, pipeline, step1, step2)
            count_steps += 1

            current_step = prev_step
            for primitive in primitives:
                current_step = make_pipeline_module(db, pipeline, primitive)
                change_default_hyperparams(db, pipeline, primitive, current_step)
                connect(db, pipeline, prev_step, current_step)
                count_steps += 1

                if primitive != primitives[-1]:  # Not update when it is the last primitive
                    prev_step = current_step

                to_module_primitive = index.get_primitive(primitive)
                if 'outputs' in to_module_primitive.metadata.query()['primitive_code']['arguments']:
                    connect(db, pipeline, step2, current_step, to_input='outputs')

            if 'ROC_AUC' in metrics[0]['metric'].name:
                add_rocauc_primitives(pipeline, current_step, prev_step, step2, dataframe_step, count_steps, db)
            else:
                step6 = make_pipeline_module(db, pipeline, 'd3m.primitives.data_transformation.'
                                                           'construct_predictions.Common')
                connect(db, pipeline, current_step, step6)
                connect(db, pipeline, step1, step6, to_input='reference')

            db.add(pipeline)
            db.commit()
            logger.info('%s PIPELINE ID: %s', origin, pipeline.id)
            return pipeline.id
        except:
            logger.exception('Error creating pipeline id=%s, primitives=%s', pipeline.id, str(primitives))
            return None
        finally:
            db.close()



import math
import numpy as np
import logging

logger = logging.getLogger(__name__)

np.random.seed(0)


class MCTS():
    """
    This class handles the MCTS tree.
    """

    def __init__(self, game, nnet, args):
        self.game = game
        self.nnet = nnet
        self.args = args
        self.Qsa = {}  # stores Q values for s,a (as defined in the paper)
        self.Nsa = {}  # stores #times edge s,a was visited
        self.Ns = {}  # stores #times board s was visited
        self.Ps = {}  # stores initial policy (returned by neural net)

        self.Es = {}  # stores game.getGameEnded ended for board s
        self.Vals = {}
        self.Vs = {}  # stores game.getValidMoves for board s
        self.count = 0

    def getActionProb(self, canonicalBoard, temp=1):
        """
        This function performs numMCTSSims simulations of MCTS starting from
        canonicalBoard.

        Returns:
            probs: a policy vector where the probability of the ith action is
                   proportional to Nsa[(s,a)]**(1./temp)
        """
        for i in range(self.args.get('numMCTSSims')):
            logger.info('MCTS SIMULATION %s', i + 1)
            self.search(canonicalBoard)

        s = self.game.stringRepresentation(canonicalBoard)
        counts = [0] * self.game.getActionSize()
        counts = [self.Nsa[(s, a)] if (s, a) in self.Nsa else 0 for a in range(self.game.getActionSize())]

        if temp == 0:
            bestA = np.argmax(counts)
            probs = [0] * len(counts)
            probs[bestA] = 1
            if np.sum(probs) == 0:
                logger.info('PROB ZERO')
            return probs

        counts = [x ** (1. / temp) for x in counts]
        if np.sum(counts) == 0:
            probs = [1 / (len(counts))] * len(counts)
        else:
            non_zero_args = list(np.where(np.asarray(counts) > 0)[0])
            probs = [0] * len(counts)
            for index in non_zero_args:
                probs[index] = counts[index] / float(sum(counts))
        if np.sum(probs) == 0:
            logger.info('PROB ZERO')
        return probs

    def search(self, canonicalBoard, player=1):
        """
        This function performs one iteration of MCTS. It is recursively called
        till a leaf node is found. The action chosen at each node is one that
        has the maximum upper confidence bound as in the paper.

        Once a leaf node is found, the neural network is called to return an
        initial policy P and a value v for the state. This value is propogated
        up the search path. In case the leaf node is a terminal state, the
        outcome is propogated up the search path. The values of Ns, Nsa, Qsa are
        updated.

        NOTE: the return values are the negative of the value of the current
        state. This is done since v is in [-1,1] and if v is the value of a
        state for the current player, then its value is -v for the other player.

        Returns:
            v: the negative of the value of the current canonicalBoard
        """
        self.game.display(canonicalBoard)

        s = self.game.stringRepresentation(canonicalBoard)
        game_ended = self.game.getGameEnded(canonicalBoard, player)
        # logger.info('GAME ENDED %s', game_ended)

        if s not in self.Es:
            self.Es[s] = game_ended
        if self.Es[s] != 0:
            # terminal node
            # Clear all previous moves
            return self.Vals[s] if not self.Vals.get(s) is None else 0

        if s not in self.Ps:
            # leaf node
            self.Ps[s], v = self.nnet.predict(self.game.getTrainBoard(canonicalBoard))
            logger.info('Prediction %s', v)
            # logger.info('CALLING VALID MOVES')
            valids = self.game.getValidMoves(canonicalBoard, 1)
            self.Ps[s] = self.Ps[s] * valids  # masking invalid moves
            self.Ps[s] /= np.sum(self.Ps[s])  # renormalize
            self.Vals[s] = v
            self.Vs[s] = valids
            self.Ns[s] = 0
            return v

        valids = self.Vs[s]

        # Check if valid moves are available. Quit if no more legal moves are possible
        if not any(valids):
            return 0

        cur_best = -float('inf')
        best_act = -1
        current_primitives = self.game.get_pipeline_primitives(canonicalBoard)
        current_pattern = ' '.join(
            [self.game.grammar['RULES_PROBA']['TYPES'].get(p, {'type': p})['type'] for p in current_primitives])
        # pick the action with the highest upper confidence bound
        actions = []
        alpha = 0.1
        if len(self.game.grammar['RULES_PROBA']['GLOBAL']) == 0:  # It's a manual grammar, so only use the NN info
            alpha = 1
        for a in range(self.game.getActionSize()):
            if valids[a]:
                # logger.info('MCTS ACTION %s', a)
                global_proba = self.game.grammar['RULES_PROBA']['GLOBAL'].get(a + 1, (0, 0))[1]
                local_proba = self.game.grammar['RULES_PROBA']['LOCAL'].get(current_pattern, {}).get(a + 1, (0, 0))[1]
                correlation = global_proba * local_proba

                if (s, a) in self.Qsa:
                    u = self.Qsa[(s, a)] + self.args.get('cpuct') * (
                                alpha * self.Ps[s][a] + (1 - alpha) * correlation) * math.sqrt(self.Ns[s]) / (
                                    1 + self.Nsa[(s, a)])
                    # u = (global_proba + local_proba) + self.Qsa[(s, a)] + self.args.get('cpuct') * math.sqrt(self.Ns[s]) / (1 + self.Nsa[(s, a)])
                    # u = self.Qsa[(s, a)] + self.args.get('cpuct') * self.Ps[s][a] * math.sqrt(self.Ns[s]) / (1 + self.Nsa[(s, a)])
                else:
                    u = self.args.get('cpuct') * (alpha * self.Ps[s][a] + (1 - alpha) * correlation) * math.sqrt(
                        self.Ns[s])  # Q = 0 ?
                    # u = (global_proba + local_proba) + self.args.get('cpuct') * math.sqrt(self.Ns[s]) # Q = 0 ?
                    # u = self.args.get('cpuct') * self.Ps[s][a] * math.sqrt(self.Ns[s])  # Q = 0 ?
                # print(self.game.grammar['RULES_PROBA']['GLOBAL'].get(a + 1, ('None', 0))[0], correlation)
                # print('u', u)
                if u > cur_best:
                    cur_best = u
                    best_act = a
                    actions = [a]
                elif u == cur_best:
                    actions.append(a)
        if len(actions) > 1:
            # print('random')
            a = np.random.choice(np.asarray(actions))
        else:
            a = best_act
        # print('>>>>>>>>>>best a=%d' % a, 'global', self.game.grammar['RULES_PROBA']['GLOBAL'].get(a + 1, ('None', 0)))
        # print('>>>>>>>>>>best a=%d' % a, 'local', self.game.grammar['RULES_PROBA']['LOCAL'].get(current_pattern, {}).get(a + 1, ('None', 0)))
        # print('>>>>>>>>>>best a=%d' % a, 'total',  self.game.grammar['RULES_PROBA']['GLOBAL'].get(a+1, (0, 0))[1] * self.game.grammar['RULES_PROBA']['LOCAL'].get(current_pattern, {}).get(a + 1, (0, 0))[1])
        # logger.info('BEST ACTIONS %s', actions)
        # logger.info('MCTS BEST ACTION %s', best_act)
        next_s, next_player = self.game.getNextState(canonicalBoard, player, a)
        next_s = self.game.getCanonicalForm(next_s, next_player)
        self.game.display(next_s)

        # logger.info('NEXT STATE SEARCH RECURSION')
        v = self.search(next_s, next_player)

        if (s, a) in self.Qsa:
            self.Qsa[(s, a)] = (self.Nsa[(s, a)] * self.Qsa[(s, a)] + v) / (self.Nsa[(s, a)] + 1)
            self.Nsa[(s, a)] += 1

        else:
            self.Qsa[(s, a)] = v
            self.Nsa[(s, a)] = 1

        self.Ns[s] += 1

        return v



import signal
import os
import sys

# Use a headless matplotlib backend
os.environ['MPLBACKEND'] = 'Agg'

from os.path import join, dirname
from alphad3m.pipeline_search.Coach import Coach
from alphad3m.pipeline_search.pipeline.PipelineGame import PipelineGame
from alphad3m.pipeline_search.pipeline.NNet import NNetWrapper
from alphad3m.grammar_loader import load_manual_grammar, load_automatic_grammar
from alphad3m.data_ingestion.data_profiler import get_privileged_data, select_encoders
from alphad3m.pipeline_synthesis.d3mpipeline_builder import *
from alphad3m.metafeature.metafeature_extractor import ComputeMetafeatures
from alphad3m.utils import get_collection_type
from d3m.metadata.problem import TaskKeyword, TaskKeywordBase


logger = logging.getLogger(__name__)


config = {
    'PROBLEM_TYPES': {'CLASSIFICATION': 1,
                      'REGRESSION': 2,
                      'CLUSTERING': 3,
                      'TIME_SERIES_FORECASTING': 4,
                      'TIME_SERIES_CLASSIFICATION': 5,
                      'COMMUNITY_DETECTION': 6,
                      'GRAPH_MATCHING': 7,
                      'COLLABORATIVE_FILTERING': 8,
                      'LINK_PREDICTION': 9,
                      'VERTEX_CLASSIFICATION': 10,
                      'OBJECT_DETECTION': 11,
                      'SEMISUPERVISED_CLASSIFICATION': 12,
                      'TEXT_CLASSIFICATION': 13,
                      'IMAGE_CLASSIFICATION': 14,
                      'AUDIO_CLASSIFICATION': 15,
                      'TEXT_REGRESSION': 16,
                      'IMAGE_REGRESSION': 17,
                      'AUDIO_REGRESSION': 18,
                      'VIDEO_CLASSIFICATION': 19,
                      'LUPI': 20,
                      'NA': 21
                      },

    'DATA_TYPES': {'TABULAR': 1,
                   'GRAPH': 2,
                   'IMAGE': 3},

    'PIPELINE_SIZE': 8,

    'ARGS': {
        'numIters': 25,
        'numEps': 5,
        'tempThreshold': 15,
        'updateThreshold': 0.6,
        'maxlenOfQueue': 200000,
        'numMCTSSims': 5,
        'arenaCompare': 40,
        'cpuct': 1,

        'checkpoint': join(os.environ.get('D3MOUTPUTDIR'), 'temp', 'nn_models'),
        'load_model': False,
        'load_folder_file': (join(os.environ.get('D3MOUTPUTDIR'), 'temp', 'nn_models'), 'best.pth.tar'),
        'metafeatures_path': '/d3m/data/metafeatures',
        'verbose': True
    }
}


def signal_handler(signal_num, frame):
    logger.info('Receiving signal %s, terminating process' % signal.Signals(signal_num).name)
    signal.alarm(0)  # Disable the alarm
    sys.exit(0)


@database.with_sessionmaker
def generate_pipelines(task_keywords, dataset, metrics, problem, targets, features, hyperparameters, metadata,
                       pipeline_template, time_bound, msg_queue, DBSession):
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(int(time_bound))

    builder = None
    task_name = 'CLASSIFICATION' if TaskKeyword.CLASSIFICATION in task_keywords else 'REGRESSION'
    # Primitives for LUPI problems are no longer available. So, just exclude privileged data
    privileged_data = get_privileged_data(problem, task_keywords)
    metadata['exclude_columns'] += privileged_data

    def eval_pipeline(primitive_names, origin):
        pipeline_id = builder.make_d3mpipeline(primitive_names, origin, dataset, pipeline_template, targets,
                                               features, metadata, metrics, DBSession=DBSession)
        # Evaluate the pipeline if syntax is correct:
        if pipeline_id:
            msg_queue.send(('eval', pipeline_id))
            return msg_queue.recv()
        else:
            return None

    if TaskKeyword.CLUSTERING in task_keywords:
        task_name = 'CLUSTERING'
        builder = BaseBuilder()
    elif TaskKeyword.SEMISUPERVISED in task_keywords:
        task_name = 'SEMISUPERVISED_CLASSIFICATION'
        builder = BaseBuilder()
    elif TaskKeyword.COLLABORATIVE_FILTERING in task_keywords:
        task_name = 'COLLABORATIVE_FILTERING'
        builder = BaseBuilder()
    elif TaskKeyword.FORECASTING in task_keywords:
        task_name = 'TIME_SERIES_FORECASTING'
        builder = BaseBuilder()
    elif TaskKeyword.COMMUNITY_DETECTION in task_keywords:
        task_name = 'COMMUNITY_DETECTION'
        builder = BaseBuilder()
    elif TaskKeyword.LINK_PREDICTION in task_keywords:
        task_name = 'LINK_PREDICTION'
        builder = BaseBuilder()
    elif TaskKeyword.OBJECT_DETECTION in task_keywords:
        task_name = 'OBJECT_DETECTION'
        builder = BaseBuilder()
    elif TaskKeyword.GRAPH_MATCHING in task_keywords:
        task_name = 'GRAPH_MATCHING'
        builder = BaseBuilder()
    elif TaskKeyword.TIME_SERIES in task_keywords and TaskKeyword.CLASSIFICATION in task_keywords:
        task_name = 'TIME_SERIES_CLASSIFICATION'
        builder = BaseBuilder()
    elif TaskKeyword.VERTEX_CLASSIFICATION in task_keywords or TaskKeyword.VERTEX_NOMINATION in task_keywords:
        task_name = 'VERTEX_CLASSIFICATION'
        builder = BaseBuilder()
    elif get_collection_type(dataset[7:]) == 'text' or TaskKeyword.TEXT in task_keywords and (
            TaskKeyword.REGRESSION in task_keywords or TaskKeyword.CLASSIFICATION in task_keywords):
        task_name = 'TEXT_' + task_name
        builder = BaseBuilder()
    elif get_collection_type(dataset[7:]) == 'image' or TaskKeyword.IMAGE in task_keywords and (
            TaskKeyword.REGRESSION in task_keywords or TaskKeyword.CLASSIFICATION in task_keywords):
        if TaskKeyword.IMAGE not in task_keywords: task_keywords.append(TaskKeyword.IMAGE)
        task_name = 'IMAGE_' + task_name
        builder = BaseBuilder()
    elif TaskKeyword.AUDIO in task_keywords and (
            TaskKeyword.REGRESSION in task_keywords or TaskKeyword.CLASSIFICATION in task_keywords):
        task_name = 'AUDIO_' + task_name
        builder = AudioBuilder()
    elif TaskKeyword.VIDEO in task_keywords and (
            TaskKeyword.REGRESSION in task_keywords or TaskKeyword.CLASSIFICATION in task_keywords):
        task_name = 'VIDEO_' + task_name
        builder = BaseBuilder()
    elif TaskKeyword.CLASSIFICATION in task_keywords or TaskKeyword.REGRESSION in task_keywords:
        builder = BaseBuilder()
    else:
        logger.warning('Task %s doesnt exist in the grammar, using default NA_TASK' % task_name)
        task_name = 'NA'
        builder = BaseBuilder()

    use_automatic_grammar = False
    include_primitives = hyperparameters.get('include_primitives', []) or []  # Use empty list when the value is None
    exclude_primitives = hyperparameters.get('exclude_primitives', []) or []  # Use empty list when the value is None

    def update_config(task_name):
        config['PROBLEM'] = task_name
        config['DATA_TYPE'] = 'TABULAR'
        config['METRIC'] = metrics[0]['metric'].name
        config['DATASET'] = problem['inputs'][0]['dataset_id']
        config['ARGS']['stepsfile'] = join(os.environ.get('D3MOUTPUTDIR'), 'temp', config['DATASET'] + '_pipeline_steps.txt')

        task_name_id = task_name + '_TASK'
        task_keywords_mapping = {v: k for k, v in TaskKeywordBase.get_map().items()}
        task_keyword_ids = [task_keywords_mapping[t] for t in task_keywords]
        grammar = None

        if use_automatic_grammar:
            dataset_path = join(dirname(dataset[7:]), 'tables', 'learningData.csv')
            target_column = problem['inputs'][0]['targets'][0]['column_name']
            grammar = load_automatic_grammar(task_name_id, dataset_path, target_column, task_keyword_ids,
                                             include_primitives, exclude_primitives)

        if grammar is None:
            encoders = select_encoders(metadata['only_attribute_types'])
            use_imputer = metadata['missing_values']
            grammar = load_manual_grammar(task_name_id, task_keyword_ids, encoders, use_imputer, include_primitives,
                                          exclude_primitives)
        config['GRAMMAR'] = grammar
        metafeatures_extractor = ComputeMetafeatures(dataset, targets, features, DBSession)
        config['DATASET_METAFEATURES'] = [0] * 50 #metafeatures_extractor.compute_metafeatures('AlphaD3M_compute_metafeatures')

        return config

    config_updated = update_config(task_name)

    game = PipelineGame(config_updated, eval_pipeline)
    nnet = NNetWrapper(game)

    if config['ARGS'].get('load_model'):
        model_file = join(config['ARGS'].get('load_folder_file')[0],
                          config['ARGS'].get('load_folder_file')[1])
        if os.path.isfile(model_file):
            nnet.load_checkpoint(config['ARGS'].get('load_folder_file')[0],
                                 config['ARGS'].get('load_folder_file')[1])

    c = Coach(game, nnet, config['ARGS'])
    c.learn()

    sys.exit(0)
